from unittest import TestCase

import pytest

from datetime import timedelta

# from eduperson.assurance import *


@pytest.mark.skip(reason='The tested code seems to be missing ?!')
class TestIdentifierUniqueness(TestCase):
    def test_empty(self):
        ass = []
        id = IdentifierUniqueness(ass)
        self.assertFalse(id.uid_uniqueness_levels)
        self.assertFalse(id.eppn_uniqueness_levels)
        self.assertFalse(id.eppn_uniqueness_reassign_period)

    def test_unique(self):
        ass = [
            "https://refeds.org/assurance/ID/unique",
        ]
        id = IdentifierUniqueness(ass)
        self.assertEqual(id.uid_uniqueness_levels, {1,2,3,4})
        self.assertFalse(id.eppn_uniqueness_levels)
        self.assertFalse(id.eppn_uniqueness_reassign_period)

    def test_no_reassign(self):
        ass = [
            "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
        ]
        id = IdentifierUniqueness(ass)
        self.assertFalse(id.uid_uniqueness_levels)
        self.assertEqual(id.eppn_uniqueness_levels, {1,2,3})
        self.assertFalse(id.eppn_uniqueness_reassign_period)

    def test_1y(self):
        ass = [
            "https://refeds.org/assurance/ID/eppn-unique-reassign-1y",
        ]
        id = IdentifierUniqueness(ass)
        self.assertFalse(id.uid_uniqueness_levels)
        self.assertEqual(id.eppn_uniqueness_levels, {1,2})
        self.assertEqual(id.eppn_uniqueness_reassign_period, timedelta(days=365))

    def test_conflict(self):
        ass = [
            "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
            "https://refeds.org/assurance/ID/eppn-unique-reassign-1y",
        ]
        with self.assertRaises(ValueError):
            IdentifierUniqueness(ass)


    def test_combine(self):
        ass = [
            "https://refeds.org/assurance/ID/unique",
            "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
        ]
        id = IdentifierUniqueness(ass)
        self.assertEqual(id.uid_uniqueness_levels, {1,2,3,4})
        self.assertEqual(id.eppn_uniqueness_levels, {1,2,3})
        self.assertFalse(id.eppn_uniqueness_reassign_period)



'''Parse commandline options'''
#
# Author: Marcus Hardt <hardt@kit.edu>
#
# pylint # {{{
# vim: tw=100 foldmethod=indent
# pylint: disable=bad-continuation, invalid-name, superfluous-parens
# pylint: disable=bad-whitespace, mixed-indentation
# pylint: disable=redefined-outer-name
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods
# }}}
import os
import sys
import logging
import argparse

logger = logging.getLogger(__name__)

def parseOptions():
    '''Parse the commandline options'''

    logger.info("reading config")
    path_of_executable      = os.path.realpath(sys.argv[0])
    folder_of_executable    = os.path.split(path_of_executable)[0]
    full_name_of_executable = os.path.split(path_of_executable)[1]
    name_of_executable      = full_name_of_executable.rstrip('.py')
    try:
        config_in_home      = os.environ['HOME']+'/.config/%s.conf' % name_of_executable
    except KeyError:
        pass

    parser = argparse.ArgumentParser(description=name_of_executable)

    parser.add_argument('--config_file', '--config','-c','--conf', 
            default='/etc/feudal/feudal_adapter.conf',
            help='Default: /etc/feudal/feudal_adapter.conf')

    return parser

# reparse args on import
args = parseOptions().parse_args()



# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
# pylint: disable=missing-docstring, too-few-public-methods
"""Results for feudalClient"""

class Result:
    """A Result returned by the adapter to the feudalClient.

    Serialized to JSON as the final act of this script.
    """

    def __init__(self, state, message):
        """Called by subclasses, usually not directly.

        Arguments:
        state -- The state that was reached. One of 'deployed', 'not_deployed', 'failure' and 'rejected' (see subclasses below).
        message -- Displayed to the user in the feudalClient webinterface.
        """
        self.state = state
        self.message = message

    @property
    def attributes(self):
        return self.__dict__


## Sucessful
class Success(Result):
    """Indicates a successful result (i.e. user was deployed or undeployed)."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

class Deployed(Success):
    """Indicates that the user was successfully deployed to the service."""
    def __init__(self, credentials, **kwargs):
        """Initialises this Result with a state of 'deployed'.

        Arguments:
        credentials -- A dictionary displayed to the user in the feudalClient webinterface.
        **kwargs -- Any additional keyword arguments are passed to Success.__init__
        """
        super().__init__(state='deployed', **kwargs)
        self.credentials = credentials

class NotDeployed(Success):
    """Indicates that the user was successfully undeployed from the service."""
    def __init__(self, **kwargs):
        super().__init__(state='not_deployed', **kwargs)

class Status(Success):
    """Indicates the status of the user."""
    def __init__(self, state, **kwargs):
        super().__init__(state=state, **kwargs)


## Exceptional (Error or Questionnaire)
class ExceptionalResult(Result, Exception):
    """Raise a subclass of this to abort and directly return this Result to the feudalClient"""
    pass

class Failure(ExceptionalResult):
    """Indicates a failure in attempting to deploy/undeploy the user.

    The previous state should be retained, but might also be inconsistent
    """
    def __init__(self, **kwargs):
        super().__init__(state='failed', **kwargs)

class Rejection(ExceptionalResult):
    """Indicates that the user is not allowed to access the requested resource.

    A reason for this might be an insufficient assurance level.
    """
    def __init__(self, **kwargs):
        super().__init__(state='rejected', **kwargs)

class Questionnaire(ExceptionalResult):
    """Additional information is needed to reach the desired state.

    Usually, this is raised during deployment to get additional
    information.  The feudalClient then calls this script again with
    an additional object under the key 'answers' in the
    input dictionary. This can be accessed from within the UserInfo
    class (see below).

    Can be used to ask multiple questions at once.

    Arguments:
    questions -- A Dictionary of `{name: text, ...}` questions. The text is the question
                 displayed to the user in the feudalClient webinterface. The answer
                 submitted by the user can then be found under the key `name` in the
                 answers dictionary.
    defaults -- A Dictionary of `{name: thing, ...}` quenstionaire answers. The `thing` is the
                default value of the named question. The type of thing determines the input of the
                answer. Possible are strings, numbers or lists to provide the user with a selection
                of the given values.
    **kwargs -- Any additional keyword arguments ar passed to ExceptionalResult.__init__
    """
    def __init__(self, questions, defaults, **kwargs):
         super().__init__(state='questionnaire', message='There are unanswered questions.', **kwargs)
         self.questionnaire = questions
         self.questionnaire_answers = defaults

class Question(Questionnaire):
    """Convenience class to ask a single question."""
    def __init__(self, name, text, default=None, **kwargs):
        """See Questionaire.__init__ for details.

        Arguments:
        name -- The name of the question
        text -- Displayed to the user
        default -- The default value / the input type
        **kwargs -- Any additional keyword arguments are passed to Questionaire.__init__
        """
        defaults = {}
        if default:
            defaults[name] = default

        super().__init__(questions={name: text}, defaults=defaults, **kwargs)

def raise_question(*args, **kwargs):
    """Convenice function needed in places where an expression is required.

    E.g: `userinfo.get_value() or raise Question(...)` is not valid,
    but `userinfo.get_value() or raise_question(...)` is.
    """
    raise Question(*args, **kwargs)




name = 'eduperson'

import logging

import regex

logger = logging.getLogger(__name__)

class Entitlement:
    """EduPerson Entitlement attribute (de-)serialisation.

    As specified in: https://aarc-project.eu/guidelines/aarc-g002/
    """

    # This regex is not compatible with stdlib 're', we need 'regex'!
    # (because of repeated captures, see https://bugs.python.org/issue7132)
    re = regex.compile(
        r'urn:' +
           r'(?P<nid>[^:]+):(?P<delegated_namespace>[^:]+)' +     # Namespace-ID and delegated URN namespace
           r'(:(?P<subnamespace>[^:]+))*?' +                      # Sub-namespaces
        r':group:' +
           r'(?P<group>[^:]+)' +                                  # Root group
           r'(:(?P<subgroup>[^:]+))*?' +                          # Sub-groups
           r'(:role=(?P<role>.+))?' +                             # Role of the user in the deepest group
        r'#(?P<group_authority>.+)'                               # Authoritative soruce of the entitlement (URN)
    )

    def __init__(self, raw):
        """Parse a raw EduPerson entitlement string in the AARC-G002 format."""
        match = self.re.fullmatch(raw)

        if not match:
            raise ValueError("Invalid entitlement attribute")

        [self.namespace_id] = match.captures('nid')
        [self.delegated_namespace] = match.captures('delegated_namespace')

        self.subnamespaces = match.captures('subnamespace')

        [self.group] = match.captures('group')
        self.subgroups = match.captures('subgroup')
        [self.role] = match.captures('role') or [None]

        [self.group_authority] = match.captures('group_authority')

    def __repr__(self):
        """Serialize the entitlement to the AARC-G002 format.

        This is the inverse to `__init__` and thus `ent_str == repr(Entitlement(ent_str))`
        holds for any valid entitlement.
        """
        return ((
            'urn:{namespace_id}:{delegated_namespace}{subnamespaces}' +
            ':group:{group}{subgroups}{role}' +
            '#{group_authority}'
        ).format(**{
            **self.__dict__, **{
                'subnamespaces': ''.join([':{}'.format(ns) for ns in self.subnamespaces]),
                'subgroups': ''.join([':{}'.format(grp) for grp in self.subgroups]),
                'role': ':role={}'.format(self.role) if self.role else ''
        }}))

    def __str__(self):
        """Return the entitlement in human-readable string form."""
        return ((
            '<Entitlement' +
            ' namespace={namespace_id}:{delegated_namespace}{subnamespaces}' +
            ' group={group}{subgroups}' +
            '{role}' +
            ' auth={group_authority}>'
        ).format(**{
            **self.__dict__, **{
                'subnamespaces': ''.join([',{}'.format(ns) for ns in self.subnamespaces]),
                'subgroups': ''.join([',{}'.format(grp) for grp in self.subgroups]),
                'role': ' role={}'.format(self.role) if self.role else ''
        }}))


    @property
    def all_toplevel_groups(self):
        return  ["_".join([self.group])]

    @property
    def all_groups(self):
        return  ["_".join([self.group] + self.subgroups)]

    @property
    def full_namespace(self):
        return [self.delegated_namespace] + self.subnamespaces




###############################################################
# all
# class User:
#  | Method                                    | unix | bwidm | ldap | Mandatory |
#  +-------------------------------------------+------+-------+------+-----------+
#  | __init__(self, userinfo)                  | y    | y     | y    |           |
#  | create(self)                              |      | y     | y    |           |
#  | credentials(self)                         |      |       |      |           |
#  | deactivate(self)                          |      | y     |      |           |
#  | delete(self)                              | y    | y     | y    |           |
#  | exists(self)                              | y    | y     | y    |           |
#  | external_user_update(self, state_updates) |      | y     |      |           |
#  | get_username(self)                        |      | y     |      |           |
#  | install_ssh_keys(self)                    | y    | y     | y    |           |
#  | is_limited(self)                          | y    |       |      |           |
#  | is_pending(self)                          | y    |       |      |           |
#  | is_rejected(self)                         | y    |       |      |           |
#  | is_suspended(self)                        | y    |       |      |           |
#  | limit(self)                               | y    |       |      |           |
#  | mod(self, supplementary_groups=None)      | y    | y     | y    |           |
#  | name_taken(self)                          | y    | y     | y    |           |
#  | reg_info(self, json=True, **kwargs)       |      | y     |      |           |
#  | resume(self)                              | y    |       |      |           |
#  | set_username(self, username)              | y    | y     |      |           |
#  | suspend(self)                             | y    |       |      |           |
#  | uninstall_ssh_keys(self)                  | y    | y     | y    |           |
#  | unlimit(self)                             | y    |       |      |           |
#  | update(self)                              | y    | y     | y    |           |
#
# class Group
#  | Method                                           | unix | bwidm | ldap | Mandatory |
#  +--------------------------------------------------+------+-------+------+-----------+
#  | __init__(self, name)                             | y    | y     | y    |           |
#  | create(self)                                     | y    | y     | y    |           |
#  | delete(self)                                     | y    | y     |      |           |
#  | exists(self)                                     | y    | y     | y    |           |
#  | mod(self)                                        | y    | y     |      |           |
#  | reg_info(self, json=True, short=False, **kwargs) |      | y     |      |           |
#
# ###############################################################
# # local_unix.py
# class User:
#     def __init__(self, userinfo):
#     def exists(self):
#     def is_rejected(self):
#     def is_suspended(self):
#     def is_pending(self):
#     def is_limited(self):
#     def name_taken(self):
#     def get_username(self):
#     def set_username(self, username):
#     def create(self):
#     def update(self):
#     def delete(self):
#     def mod(self, supplementary_groups=None):
#     def suspend(self):
#     def resume(self):
#     def limit(self):
#     def unlimit(self):
#     def install_ssh_keys(self):
#     def uninstall_ssh_keys(self):
# class Group:
#     def __init__(self, name):
#     def exists(self):
#     def create(self):
#     def delete(self):
#     def mod(self):
#
# ###############################################################
# # bwidm.py
# class BwIdmConnection:
# class User:
#     def __init__(self, userinfo):
#     def exists(self):
#     def name_taken(self):
#     def get_username(self):
#     def set_username(self, username):
#     def create(self):
#     def update(self):
#     def delete(self):
#     def deactivate(self):
#     def mod(self, supplementary_groups=None):
#     def install_ssh_keys(self):
#     def uninstall_ssh_keys(self):
#     def external_user_update(self, state_updates):
#     def reg_info(self, json=True, **kwargs):
# class Group:
#     def __init__(self, name):
#     def exists(self):
#     def create(self):
#     def delete(self):
#     def mod(self):
#     def reg_info(self, json=True, short=False, **kwargs):
#
#     @property
#     def members(self):
#         raise NotImplementedError('Do we even need this function?')
#
# ###############################################################
# # ldap.py
# """Simple example backend, for documentation purposes only.
# It's in the proof-of-concept state.
# """
# class User:
#     def __init__(self, userinfo):
#     def exists(self):
#     def name_taken(self):
#     def create(self):
#     def update(self):
#     def delete(self):
#     def mod(self, supplementary_groups=None):
#     def install_ssh_keys(self):
#     def uninstall_ssh_keys(self):
#
#     @property
#     def credentials(self):
#
# class Group:
#     """Manages the group object on the service."""
#     def __init__(self, name):
#     def exists(self):
#     def create(self):
#
#



# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
# pylint: disable=missing-docstring, too-few-public-methods
"""
BWIDM backend.

See https://git.scc.kit.edu/simon/reg-app.
"""

import logging
import json
from functools import reduce
from time import sleep

import requests
import os

from urllib.parse import urljoin

from ..config import CONFIG
from .. import utils
from time import sleep

from ldf_adapter.logsetup import jsonlogger
logger = logging.getLogger(__name__)

class BwIdmConnection:
    """Connection to the BWIDM API."""
    def __init__(self, config=None):
        self.session = requests.Session()
        if config:
            self.session.auth = (
                config['backend.bwidm.auth']['http_user'],
                config['backend.bwidm.auth']['http_pass']
            )

        if not CONFIG['backend.bwidm'].getboolean('log_outgoing_http_requests', fallback=False):
            logging.getLogger("requests").setLevel(logging.CRITICAL)
            logging.getLogger("werkzeug").setLevel(logging.CRITICAL)
            logging.getLogger("urllib3").setLevel(logging.CRITICAL)

    def get(self, *url_fragments, **kwargs):
        return self._request('GET', url_fragments, **kwargs)

    def post(self, *url_fragments, **kwargs):
        return self._request('POST', url_fragments, **kwargs)

    def _request(self, method, url_fragments, **kwargs):
        """
        Arguments:
        method -- HTTP Method (type: str)
        url_fragments -- The components of the URL. Each is url-encoded separately and then they are
                         joined with '/'
        fail=True -- Raise exception on non-200 HTTP status
        **kwargs -- Passed to `requests.Request.__init__`
        """
        fail = kwargs.pop('fail', True)

        url_fragments = map(str, url_fragments)
        url_fragments = map(lambda frag: requests.utils.quote(frag, safe=''), url_fragments)
        url = reduce(lambda acc, frag: urljoin(acc, frag) if acc.endswith('/') else urljoin(acc+'/', frag),
                     url_fragments,
                     CONFIG['backend.bwidm']['url'])

        req = requests.Request(method, url, **kwargs)
        rsp = self.session.send(self.session.prepare_request(req))

        if fail:
            if not rsp.ok:
                logger.error("Server responded with: {}".format(rsp.content.decode('utf-8')))
            rsp.raise_for_status()

        return rsp

BWIDM = BwIdmConnection(CONFIG)

class User:
    ATTR_USERNAME = 'urn:oid:0.9.2342.19200300.100.1.1'
    ATTR_ORG_ID = 'http://bwidm.de/bwidmOrgId'
    VALUE_USER_ACTIVE = 'ACTIVE'
    VALUE_USER_INACTIVE = 'ON_HOLD'

    def __init__(self, userinfo):
        self.info = userinfo
        self.credentials = {}
        self.primary_group = Group(userinfo.primary_group)
        self.force_username = None

    def exists(self):
        """
        Inactive users ('ON_HOLD') are treated as nonexistent.
        """
        return self._exists() and self._is_active()

    def _exists(self):
        exists = b'no such user' not in self.reg_info(json=False, fail=False)
        logger.debug('User {} {} on service BWIDM'.format(
            self.info.unique_id, 'exists' if exists else "doesn't exist"
        ))
        return exists

    def _is_active(self):
        status = self.reg_info()['userStatus']
        logger.debug('User {} is {} on service BWIDM'.format(
            self.info.unique_id, status
        ))
        return status == self.VALUE_USER_ACTIVE
    def _is_registered(self):
        """
        find if the user is already registered for a given
        service, identified by its service short name
        """
        # FIXME: Consider putting this request into the global user object (to reduce load on regapp)
        ssn = CONFIG['backend.bwidm.service']['name']
        registrations = BWIDM.get ('external-reg', 'find', 'externalId',
                                    self.info.unique_id)
        # find registrations
        number_of_registrations = 0
        try:
            logger.debug("logging registrations to jsonlog")
            jsonlogger.debug(json.dumps(registrations.json, sort_keys=True, indent=4, separators=(',', ': ')))
        except TypeError:
            pass
        for reg in registrations.json():
            if reg["serviceShortName"] == ssn:
                if reg["registryStatus"] == "ACTIVE":
                    number_of_registrations += 1
        logger.debug(F"Number of registrations found: {number_of_registrations}")
        if number_of_registrations > 0:
            return True
        return False

    def name_taken(self, name):
        """
        If there is a user for our unique_id with our username, treat the name as available. This
        might happen if the our user is ON_HOLD on the service.
        """
        # TODO: argument "name" was added, check if given "name" (instead of info.username) is taken by *another* user
        logger.debug(F"checking for name taken of {name}")
        users_with_name = BWIDM.get(
            'external-user', 'find',
            'attribute', self.ATTR_USERNAME, name
        ).json()

        other_users_with_name = [user for user in users_with_name if user['externalId'] != self.info.unique_id]
        # logger.debug (F"other_users: {other_users_with_name}")
        logger.debug(F"Found {len(other_users_with_name)} with same username")
        # FIXME: Why < ?
        if len(other_users_with_name) < len(users_with_name):
            logger.debug("Username '{}' is reserved for us".format(self.info.username))

        if other_users_with_name:
            logger.error("Username '{}' is already used by\n    {}".format(
                self.info.username,
                ",\n    ".join(map(lambda u: u['externalId'], other_users_with_name))
            ))
        else:
            logger.debug("Username '{}' is available".format(name))

        return bool(other_users_with_name)

    def get_username(self):
        """Check if a user exists based on unique_id and return the name"""
        def safe_resp_conversion(resp):
            '''Safely convert a response to json'''
            if resp.status_code != 200:
                logger.debug ('Error %d reading from remote: \n%s\n'% (resp.status_code, str(resp.text)))
                s_exit(1) # or raise or return None?
            try:
                resp_json = resp.json()
            except json.JSONDecodeError:
                logging.error ('Could not decode json that I obtained from rest server')
                raise
            return resp_json
        full_username = None
        resp          = BWIDM.get ('external-user', 'find', 'externalId', self.info.unique_id)
        resp_json     = safe_resp_conversion(resp)

        try:
            username = resp_json['attributeStore']['urn:oid:0.9.2342.19200300.100.1.1']
            bwIdmOrgId = resp_json['attributeStore']['http://bwidm.de/bwidmOrgId']
            full_username = F"{bwIdmOrgId}_{username}"
        except KeyError as e:
            logger.error('Error: I could not find the username in the database.')
            logger.error('  Most likely the user is not registered for this service\n')
            logger.error(F"  {e}")
            logger.error(json.dumps(resp_json, sort_keys=True, indent=4, separators=(',', ': ')))
        logger.debug(F"Found existing username: {full_username}")
        return full_username

    def set_username(self, username):
        """Update the internal representation of the user with the incoming username"""
        logger.debug("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
        self.force_username = username
        logger.debug("bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb")

    def create(self):
        """Create or activate user."""
        if self._exists() and not self._is_active():
            logger.info("Activating user {unique_id}".format(**self.info))
            BWIDM.get('external-user', 'activate', 'externalId', self.info.unique_id)
        else:
            logger.info("Creating user {unique_id}".format(**self.info))
            BWIDM.post('external-user', 'create', json={
                'externalId': self.info.unique_id
            })

    def update(self):
        # FIXME: Consider putting this request into the global user object (to reduce load on regapp)
        def get_active_reg_info(ext_id):
            rsp = BWIDM.get('external-reg', 'find',
                            'externalId', ext_id)

            try:
                return next(filter(lambda reg: reg['registryStatus'] == "ACTIVE", rsp.json()))
            except StopIteration:
                return {'lastReconcile': None}

        self.external_user_update({
            'externalId': self.info.unique_id,
            'eppn': self.info.eppn,
            'email': self.info.email,
            'givenName': self.info.given_name,
            'surName': self.info.family_name,
            'primaryGroup': {
                'id': self.primary_group.reg_info()['id']
            },
            'attributeStore': {
                self.ATTR_USERNAME: self.force_username or self.info.username,
                self.ATTR_ORG_ID: CONFIG['backend.bwidm']['org_id'],
            }
        })
        # if the user is already registered for the service, we're done here
        if self._is_registered():
            return

        old_reg = get_active_reg_info(self.info.unique_id)
        # We wait until the 'lastReconciled' timestamp changes, which means that our update was sucessfully deployed
        reg = old_reg
        while reg['lastReconcile'] == old_reg['lastReconcile']:
            sleep(0.3)
            logger.debug("Received registration reconciled at {}. That is not up-to-date. Checking again.".format(
                reg['lastReconcile']))

            rsp = BWIDM.get('external-reg', 'register',
                            'externalId', self.info.unique_id,
                            'ssn', CONFIG['backend.bwidm.service']['name'])

            if rsp.status_code == 204:
                reg = get_active_reg_info(self.info.unique_id)
            else:
                reg = rsp.json()

        logger.debug("Registration confirmed reconciled at {}. Looks like the update went through.".format(reg['lastReconcile']))

        self.credentials['ssh_user'] = reg['registryValues']['localUid']
        self.credentials['ssh_host'] = CONFIG['backend.bwidm.login_info'].get('ssh_host', 'undefined')
        self.credentials['commandline'] = "ssh {}@{}".format(
            self.credentials['ssh_user'], self.credentials['ssh_host'])

    def delete(self):
        """Deregister the user from the given service in BWIDM."""
        BWIDM.get('external-reg', 'deregister', 'externalId', self.info.unique_id,\
                  'ssn', CONFIG['backend.bwidm.service']['name'])

    def deactivate(self):
        """Deactivate the user, this does not delete from BWIDM, but sets
        the status to ON_HOLD, thereby disabling the user from ALL
        services."""
        BWIDM.get('external-user', 'deactivate', 'externalId', self.info.unique_id)

    def mod(self, supplementary_groups=None):
        reg_info = self.reg_info()

        if supplementary_groups is not None:
            current_groups = [grp for grp in reg_info['secondaryGroups']]
            new_groups = [grp.reg_info(short=True) for grp in supplementary_groups]
            new_groups += [self.primary_group.reg_info()]

            NL='\n    '
            logger.debug(F"Groups according to BWIDM: {NL}{NL.join([g['name'] for g in current_groups])}")
            logger.debug(F"Groups according to FEUDAL: {NL}{NL.join([g['name'] for g in new_groups])}")

            # Remove user from groups he should not be a member of
            to_be_removed_from = [g for g in current_groups
                                  if g['id'] not in (ng['id'] for ng in new_groups)]

            # Only add user to groups she is not already a member of
            to_be_added_to = [g for g in new_groups
                              if g['id'] not in (cg['id'] for cg in current_groups)]

            if to_be_removed_from:
                logger.info('Remove user {} from groups {}'.format(
                    self.info.username, ",".join(g['name'] for g in to_be_removed_from)))
            for grp in to_be_removed_from:
                BWIDM.get('group-admin', 'remove', 'groupId', grp['id'], 'userId', reg_info['id'])

            if to_be_added_to:
                logger.info('Add user {} to groups {}'.format(
                    self.info.username, ",".join(g['name'] for g in to_be_added_to)))
            grp_add_retvals = []
            for grp in to_be_added_to:
                grp_add_retvals.append(BWIDM.get('group-admin', 'add', 'groupId', grp['id'], 'userId', reg_info['id'], fail=False))
            if len(grp_add_retvals) > 0:
                logger.debug(F"Group add retvals: {grp_add_retvals}")

    def install_ssh_keys(self):
        self.external_user_update({
            'externalId': self.info.unique_id,
            'genericStore': {
                **self.reg_info()['genericStore'],
                'ssh_key': json.dumps(self.info.ssh_keys)
            }
        })

    def uninstall_ssh_keys(self):
        """Uninstall any SSH keys stored in BWIDM."""
        self.external_user_update({
            'externalId': self.info.unique_id,
            'genericStore': {'ssh_key': None}
        })

    def external_user_update(self, state_updates):
        """Apply new attributes to the user, performing sensible merging of dicts.

        BWIDM is a bit weird about this, due to technical restrictions in the Java-Software stack.

        This comes down to:
        {..., k: val, ...} means `state[k] = val`
        {..., k: None, ...} means `del state[k]` or `state[k]=None`
        {..., k: {}, ...} means no change to k
        {..., k: val={...}, ...} means `state[k]=merge state[k] with val`

        This is applied recursivly.

        """
        current_state = self.reg_info()
        new_state = utils.dictmerge(current_state, state_updates)
        utils.log_dictdiff(utils.dictdiff(current_state, new_state),
                           log_function=logger.info)

        for k in list(new_state):
            if new_state[k] is None:
                new_state[k] = {}

        try:
            formatted_json = (json.dumps(state_updates, sort_keys=True, indent=4, separators=(',', ': ')))
            logger.debug("  logging state_updates to jsonlog")
            jsonlogger.debug(F"state_updates:  {formatted_json}")

            formatted_json = (json.dumps(current_state, sort_keys=True, indent=4, separators=(',', ': ')))
            logger.debug("  logging current_state_updates to jsonlog")
            jsonlogger.debug(F"current_state: {formatted_json}")

            logger.debug("  logging new_to jsonlog")
            formatted_json = (json.dumps(new_state, sort_keys=True, indent=4, separators=(',', ': ')))
            jsonlogger.debug(F"new state for regapp:  {formatted_json}")
        except:
            pass
        BWIDM.post('external-user', 'update', json=new_state)

    def reg_info(self, json=True, **kwargs):
        # FIXME: Cache this functions results!
        rsp = BWIDM.get('external-user', 'find', 'externalId', self.info.unique_id, **kwargs)
        return rsp.json() if json else rsp.content

class Group:
    def __init__(self, name):
        self.name = name

    def exists(self):
        # FIXME: Group existence needs to be checked with using also
        # CONFIG['backend.bwidm.service']['name']
        return b'no such group' not in BWIDM.get('group-admin', 'find', 'name', self.name, fail=False).content

    def create(self):
        rsp = BWIDM.get('group-admin', 'create', CONFIG['backend.bwidm.service']['name'], self.name).json()

        if self.name != rsp['name']:
            logger.warning("Groupname changed from {} to {} by BWIDM".format(self.name, rsp['name']))
            self.name = rsp['name']

        self.id = rsp['id']

    def delete(self):
        # groupdel
        raise NotImplementedError('Do we even need this function?')

    def mod(self):
        # groupmod
        raise NotImplementedError('Do we even need this function?')

    def reg_info(self, json=True, short=False, **kwargs):
        rsp = BWIDM.get('group-admin', 'find' if short else 'find-detail', 'name', self.name, **kwargs)
        return rsp.json() if json else rsp.content

    @property
    def members(self):
        raise NotImplementedError('Do we even need this function?')





import logging

from ldf_adapter import *

from ldf_adapter.config import CONFIG

logger = logging.getLogger(__name__)

def test_backend(backend, user_data):
    # TODO This is not a real test
    backend.User(UserInfo(user_data))



import sys
import pkgutil
import logging

from ..config import CONFIG
from ..results import Failure

logger = logging.getLogger(__name__)

__backend__ = None
try:
    __backend__ = f"{__name__}.{CONFIG['ldf_adapter']['backend']}"
except KeyError: # Meaning: We did _NOT_ read any config, but we are used
                 # as a library
    logger.error("No configuration found; Cannot load backend")
if not __backend__:
    raise Failure(message="No configuration found; Cannot load backend")

__import__(__backend__)

class Backend:
    """Need to put __getattr__ in a class for Python < 3.7 compatibility.

    This class acts as a module.

    See https://stackoverflow.com/questions/2447353/getattr-on-a-module#7668273
    """
    __backend__ = CONFIG['ldf_adapter']['backend']

    def __getattr__(self, name):
        """Return the `User` and `Group` from the configured backend."""
        if name in ['User', 'Group']:
            return getattr(sys.modules[f"{__name__}.{self.__backend__}"], name)
        else:
            raise AttributeError(f"backend module '{__name__}' has no attribute '{name}'")

# Replace this very module (`ldf_adapter.backend`) with the pseudo-module above, which has the
# effect of this ldf_adapter.backend acting as the actually configured backend module.
sys.modules[__name__] = Backend()




from unittest import TestCase

from itertools import repeat

from ldf_adapter import *

logger = logging.getLogger(__name__)


class UserInfoTest(TestCase):

    def test_sub_masked_for_bwidm_eppn_unchanged(self):

        subs = [
            "MWMQb4ybpHVSThMGpRKkqFDJIYlGLXl1CWXSRgM8bQGR9mMXRXtMbLFubL8Sua6vZn8Dq9X3YGoKR",
            "bNAkgXeaN2rlP83UeckV0fSjU2qNmKjQ7BsOsGFC7KB1PHtYGxRXkdSZ6S1egB085cwkIYt0NNPe",
            "0Y7zkfbvLmFVclSBFqtioE0xAaV4ZtMJ7tHxScwAN6FKoPn9R3aSXjoqp1jxRWFnyN7kNoq0nvJ33f7",
            "CO6bVYGIPtzPMHpm9o13S1bztxv6jHVEAXsGX3yBRmSv8RNnVsyzjJ67bvuvl4Tq2T8rPNTcfRRqgFF",
            "X5MOmSBpHzNj3h2BzHwhCFnHzMEpT7BxJEh6X0sgpqp7TBuUdLnfBIoovpQGEYZ7zxloVgmBV5B0FUa",
            "QHp9whOAuvUgmbJBUas1haLQfa4y4VnRlxbJwqm7pqvOfZudZAbDCXlIi20ifqfBYyjQpFN9XbOUSHi",
            "bQZO9gf6K08rlRilSYNlIuroaUBvlTV2QmHFHUlm8HUWZHs8RlNnvxQZXZZ62Oju3GARJQTx8Vb8vkF",
            "YqZE73sSxnOgoJ4jW0AhFfyZZPoaQGuzWZljOL4Nmt4x6VITfrH8EzJNES2uc0mOfGZrEIWUrxCHD6J",
            "LrcvjVB6q3omCRjzoX2uLJZzurDcU0srxJqoSmOSdCvdxisxvRo9nj52jCUxYCEnEW3FV4fKFt25Cqq4",
            "p32UOxqQGTh7xD8mfY1w9XVc0j2II06oXjw19iGhWUDMOMJiOFbwbwQRR6XVyHDwnR5hPA4bWlAxi6SK",
            "TRJvjHpnkwdsKpCYyjOPbbk5Zvk5YDfaO0BcRXpyntpuUQg5WLu9esrkqxHxH8o28A6An2bSG8mIP",
            "wl1dBQBFgPHFgetwz9m6VOYeTzGn2Pu3AMDtPRX6V0XvzhxhL7y9iEatlPqiSl5OO0yiSQWDz9GJ4Ut",
            "wTAg6ewnKtcpdj1t83NWojg7dr7ydjMhZf93hLCR5NA4GtlVfAJGwmy5XGO0AVMWMO0jhAw2KMhVK5t",
            "IkYQVi17paCZdkURaFput1p5FA2rCemT8szPIwtlDnbeecVuzEdsYVT7C3zXL8TTPKQXDdMxKYbibWO5",
            "yp1Xe4fhSGwdEyqzeejRwL8wxVeiMmamjcydJmR5hjIbQ5AQoN7pe4PwGSvNKXvkQmUO09ju0gVFALh",
            "QP8b8AwjfCw7n20WAyOYDCLztANZaKvc4uCyVtYwJnFWHYaXqFLlolPc9LJ7bb7shaRvFDpQHt07Qo",
            "nScpPJSkxQja7eMtgjM83MILbYarScUR4PWlLXK40rLHavl4dDty2OV6QJLH8AVs7LMtaOrWieHAiY",
            "UuTbVNi2RknFwKfy2n3XfCqkrscrIi0cdpJXVrbzIxBqgcvOOCWGL9YFXVTjWhbFjCvzxFgTR22yKw0X",
            "Vb15x73G3eYqtCEaqjM2MaGtUmKxNansoJxPhff5wtIK1VhDO6PutZrYxEAsxoZYIrhIZMrZpc926h",
            "w72mv9VQ62SDcBGFw4Izw7Vj5eH0JYRXdN0xImmIfZtAszqFz5mrB5aBxTZKqieTExDJWlrlAGfHIAZ"
        ]

        for sub in subs:
            info = UserInfo({'user': {'userinfo': {'sub': sub}}})
            self.assertEqual(info._sub_masked_for_bwidm_eppn(), sub)

    def test_sub_masked_for_bwidm_eppn_rnd(self):
        # Generated with: cat /dev/urandom | tr -dc '[:graph:]' | tr -d 'a-zA-Z0-9_!#$%&*+/=?{|}~^.\-' | fold -w 80 | head -n 20 | sed 's/\\/\\\\/g;s/'\''/\\'\''/g;s/^/'\''/;s/$/'\'',/'

        subs = [
            ')@\\\'>""\'`,,();@"",,[];>,@,((::;;\'\']<)[:]"@;<\\<",])]>\'][@`<[\\<)>();("`"@\';;[(;[@;',
            ',`@:(,(>>;,()],["(@,\'"<]:[[,::;,@>`<[;\\:,@,(",:;)]`::;;>@<](><:(],@]\\:\\\'\'>>:\\\\>"',
            '@;<@,)]`[],:,,;\\>,](":,]@"@\'>)>]]@(]```(\\(),"(\\[,"(<`<[):;>`(:`,>,\':<[>`";>),(@@',
            '<`:;>``,@";))[\']@((>:)>`@\'":)>:"`,)><,\\""`,\';)<`\\>]`\'((\\>("<]\\\\\'>\\>\\\\"">],\\[<\'[\\',
            '\'\'\\\\\'>><\'>]\'`\'<)>,(;:,`)\'>\'];,,]@@";]]\',[\\([;`([>@[\\:;)@(<[@"`,`@\'\'"));,\\><`>>,@',
            '(`>@]`)\':]>@""@@<:;::,>\\\\(":`<`]\'(<@\')[\\\'(\'@\\:";"`(<[:][[@<,"),<\\,;));\'>`\'\\>\':[)',
            '(]`\\"\\`\'(;;)\\[\\)`\\[`)](>`]`)\\[]\\:>::@;\\[\',[><`:">@@,(@\'`[@@````,`<,\'])(\\[]\'(]<[<',
            ']<[\'`)<`[)>\\<]\']@>,(""["<,":](""\'>,,\'@\\]><)\\\',(\\"("[``\'@];<"(:>[>\')\\;)::,[:;>",[',
            '"(()\'\'\\\'@)]:;:;\'<":])`\\(`\'(](),@;:]\'>);";\'\\((::[;><)\'(<[`(::(>>(@)(;\\,@[`]):,`,)',
            '`)<"\']@:@):"]`\\);:\\\\`">)\\@>><;>)],<:;\\],`<[]])(\'\\\\,(,\\:`(@>;\'[))<,,"];@[`,<>\\\\\';',
            '(>`";>\'\'@;)")\'\\;<\\`<:<,"]@:(:)">\\@\\<[;)[:<@"\\(],(":)<\'"<,<`>:)@\';@`]:,[\\\\@<[\'>>\\',
            '])``\\;"]`):\\`),):\';\';[\'")>,`[<">":>>(\'](,[)<\'@;:[[@@]`@;((@<<;`\\():;[:<,\'`>>\\[",',
            ']]"<@``;;<<]);<];):\'[,<<>\\>))>(>`)")\'(`<]>:@<:;,@>)\';(:)>))\'\';::>[]<]\'`)@["<`<](',
            ')<>][`";["",():`]@@[`];(;)\'\\\':,]`[\\[@]\'"";":<",[)())\\<,];"<,\'"<,:]"::(]<>@](@)<\\',
            '<;::>\'];(>[;;],]:(@<<:",<\'>\\@,""`(@\\\\@)@\'<[`,:<(\'`;<@\\"><@>:[](,`<]""`@"[""[)(\\:',
            '::>"(:(())[\\\'(,[\'((`):\']\';`"]`@,\';:];@<`<:\\,;:">:)<\'@(()]\\"<;"(>[());[@@:][:;,]`',
            ';``@[,,[@<;;,]<[";\'`["":;<[(;:])[;@\'\'>@<():`)""<"<\\,@><@@)\']<@"@`]\'(@>(<@@@\'`\'":',
            '@\'[<,[>:@@[`(;:[<<](:@)<<,>">"[\\,:\'@\']\';,\\\\"(]:<,`<",>],\'>`:[,)(@([:>\\\\@]):@\\;,>',
            '))"]["(@,[[\']]<);)\'<)`@,@:"`\'`">"<[\'])\';;("]:,"\\(]@>@\'`;:()(`>)>[]>)<<,>:;,:;@[;',
            ':)]](,);\\\'(\\,,:\\`(<`<\'[@(,;[;<""""`<<(,@@",)`><@[)\'`"`]<]<\\\\:(`\\`>`<;`;(["\\,[[[]',
        ]

        for sub in subs:
            info = UserInfo({'user': {'userinfo': {'sub': sub}}})
            self.assertEqual(info._sub_masked_for_bwidm_eppn(), "".join(repeat('-', 80)))


    def test_iss_masked_for_bwidm_eppn_fixes_prefix(self):
        isss = [
            "example.org",
            "http://example.org",
            "https://example.org"
        ]

        for iss in isss:
            info = UserInfo({'user': {'userinfo': {'iss': iss}}})
            self.assertEqual(info._iss_masked_for_bwidm_eppn(), "example.org")

    def test_iss_masked_for_bwidm_eppn_fixes_umlaute(self):
        isss = [
            ("exämple.org","example.org"),
            ("example.örg","example.org"),
            ("ürsula.org","ursula.org"),
        ]

        for raw,cooked in isss:
            info = UserInfo({'user': {'userinfo': {'iss': raw}}})
            self.assertEqual(info._iss_masked_for_bwidm_eppn(), cooked)

    def test_iss_masked_for_bwidm_eppn_fixes_urls(self):
        isss = [
            ("example.org/foobar","example.org-foobar"),
            ("example.org/foo%20bar","example.org-foo-20bar"),
        ]

        for raw,cooked in isss:
            info = UserInfo({'user': {'userinfo': {'iss': raw}}})
            self.assertEqual(info._iss_masked_for_bwidm_eppn(), cooked)


    def test_group_masked_for_bwidm_converts_camel_to_snake_case(self):
        isss = [
            ("fooBarBaz", "foo_bar_baz"),
            ("FooBarBaz", "foo_bar_baz"),
        ]

        for raw,cooked in isss:
            info = UserInfo({'user': {'userinfo': {}}})
            self.assertEqual(info._group_masked_for_bwidm(raw), cooked)

    def test_group_masked_for_bwidm_all_caps(self):
        isss = [
            ("FOOBARBAZ", "foobarbaz"),
            ("FOO-BAR-BAZ", "foo-bar-baz"),
        ]

        for raw,cooked in isss:
            info = UserInfo({'user': {'userinfo': {}}})
            self.assertEqual(info._group_masked_for_bwidm(raw), cooked)

    def test_group_masked_for_bwidm_fixes_beginning(self):
        isss = [
            ("42", "four_2"),
            ("--test--", "test--"),
            ("__init__()", "init__--"),
            ("?!#_bullshit", "bullshit")
        ]

        for raw,cooked in isss:
            info = UserInfo({'user': {'userinfo': {}}})
            self.assertEqual(info._group_masked_for_bwidm(raw), cooked)


class UserInfoIntegration(TestCase):
    def assert_name(self, info):
        self.assertEqual(info.given_name, "Marcus")
        self.assertEqual(info.family_name, "Hardt")
        self.assertEqual(info.full_name, "Marcus Hardt")

    def test_unity(self):
        input = {
            "display_name": "Marcus Hardt",
            "eduperson_assurance": [
                "https://refeds.org/assurance/IAP/medium",
                "https://refeds.org/assurance/IAP/local-enterprise",
                "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
                "https://refeds.org/assurance/ATP/ePA-1m",
                "https://refeds.org/assurance/ATP/ePA-1d",
                "https://refeds.org/assurance/ID/unique",
                "https://refeds.org/assurance/profile/cappuccino",
                "https://refeds.org/assurance/IAP/low"
            ],
            "eduperson_entitlement": [
                "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
                "urn:geant:h-df.de:group:MyExampleColab#login.helmholtz-data-federation.de",
                "urn:geant:h-df.de:group:wlcg-test#login.helmholtz-data-federation.de",
                "urn:geant:h-df.de:group:HDF#login.helmholtz-data-federation.de"
            ],
            "eduperson_principal_name": "lo0018@kit.edu",
            "eduperson_scoped_affiliation": "member@kit.edu",
            "eduperson_unique_id": "6c611e2a2c1c487f9948c058a36c8f0e@login.helmholtz-data-federation.de",
            "email": "marcus.hardt@kit.edu",
            "email_verified": "true",
            "family_name": "Hardt",
            "given_name": "Marcus",
            "groups": [
                "/wlcg-test",
                "/IMK-TRO-EWCC",
                "/MyExampleColab",
                "/HDF",
                "/"
            ],
            "name": "Marcus Hardt",
            "preferred_username": "marcus",
            "sn": "Hardt",
            "ssh_key": "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIAqA5FW6m3FbFhCOsRQBxKMRki5qJxoNhZdaeLXg6ym/ marcus@nemo2019\n",
            "sub": "6c611e2a-2c1c-487f-9948-c058a36c8f0e"
        }

        info = UserInfo({'user': {'userinfo': input}})
        self.assert_name(info)
        self.assertEqual(info.username, "marcus")
        self.assertEqual(info.email, "marcus.hardt@kit.edu")
        self.assertEqual(sorted(info.groups), sorted(["imk-tro-ewcc", "my_example_colab", "wlcg-test", "hdf"]))

        self.assertEqual(info.unique_id,
                         "6c611e2a2c1c487f9948c058a36c8f0e@login.helmholtz-data-federation.de")
        self.assertEqual(info.eppn, "lo0018@kit.edu")

        self.assertTrue(info.assurance.profile.is_cappuccino())


    input_egi = {
        "acr": "https://aai.egi.eu/LoA#Substantial",
        "eduperson_assurance": [
            "https://aai.egi.eu/LoA#Substantial"
        ],
        "email": "marcus.hardt@kit.edu",
        "family_name": "Hardt",
        "given_name": "Marcus",
        "preferred_username": "mhardt",
        "sub": "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu",
        "iss": "egi.eu"
    }

    def test_egi(self):
        info = UserInfo({'user': {'userinfo': self.input_egi}})
        self.assert_name(info)
        self.assertEqual(info.username, "mhardt")
        self.assertEqual(info.email, "marcus.hardt@kit.edu")

    def test_egi_id(self):
        info = UserInfo({'user': {'userinfo': self.input_egi}})
        self.assertEqual(info.unique_id, "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu")
        self.assertEqual(info.eppn, info.unique_id)

    def test_egi_sub_is_unscoped(self):
        info = UserInfo({'user': {'userinfo': self.input_egi}})
        self.assertIn('@', info.userinfo['sub'])


    input_deep_iam = {
        "external_authn": {
            "iss": "https://accounts.google.com",
            "sub": "104223951181002749851",
            "type": "oidc"
        },
        "family_name": "Hardt",
        "given_name": "Marcus",
        "groups": [
            "KIT-Cloud"
        ],
        "name": "Marcus Hardt",
        "organisation_name": "deep-hdc",
        "preferred_username": "marcus",
        "sub": "d9730f60-3b19-4f45-83ab-f29addf72d58",
        "updated_at": "Mon Jun 25 16:55:15 CEST 2018"
    }

    def test_deep_iam(self):
        info = UserInfo({'user': {'userinfo': self.input_deep_iam}})
        self.assert_name(info)
        self.assertEqual(info.username, "marcus")
        self.assertEqual(sorted(info.groups), []) # We only support entitlements

    def test_deep_iam_id(self):
        info = UserInfo({'user': {'userinfo': self.input_egi}})
        with self.assertRaises(KeyError):
            self.assertEqual(info.unique_id, "d9730f60-3b19-4f45-83ab-f29addf72d58@deep-hdc")


    input_indigo_iam = {
        "external_authn": {
            "iss": "https://accounts.google.com",
            "sub": "104223951181002749851",
            "type": "oidc"
        },
        "family_name": "Hardt",
        "gender": "M",
        "given_name": "Marcus",
        "groups": [
            "Users",
            "Developers",
            "test.vo-users"
        ],
        "name": "Marcus Hardt",
        "organisation_name": "indigo-dc",
        "preferred_username": "marcus",
        "sub": "a1ea3aa2-8daf-41bb-b4fb-eb88f439e446",
        "updated_at": 1563283972
    }

    def test_indigo_iam(self):
        info = UserInfo({'user': {'userinfo': self.input_indigo_iam}})
        self.assert_name(info)
        self.assertEqual(info.username, "marcus")
        self.assertEqual(sorted(info.groups), sorted(["users", "developers", "test-vo-users"]))

    def test_indigo_iam_id(self):
        info = UserInfo({'user': {'userinfo': self.input_indigo_iam}})
        self.assertEqual(info.unique_id, "d9730f60-3b19-4f45-83ab-f29addf72d58@indigo-dc")
        self.assertEqual(info.eppn, info.unique_id)


    input_kit = {
        "displayName": "Hardt, Marcus (SCC)",
        "eduPersonEntitlement": [
            "urn:geant:kit.edu:group:DFN-SLCS",
            "urn:geant:kit.edu:group:LSDF-DIS",
            "urn:geant:kit.edu:group:bwGrid",
            "urn:geant:kit.edu:group:bwLSDF-FS",
            "urn:geant:kit.edu:group:bwUniCluster",
            "urn:geant:kit.edu:group:bwsyncnshare",
            "urn:geant:kit.edu:group:bwsyncnshare-idm",
            "urn:geant:kit.edu:group:gruppenverwalter"
        ],
        "eduPersonPrincipalName": "lo0018@kit.edu",
        "eduPersonScopedAffiliation": [
            "employee@kit.edu",
            "member@kit.edu"
        ],
        "eduperson_entitlement": [
            "urn:geant:kit.edu:group:DFN-SLCS",
            "urn:geant:kit.edu:group:LSDF-DIS",
            "urn:geant:kit.edu:group:bwGrid",
            "urn:geant:kit.edu:group:bwLSDF-FS",
            "urn:geant:kit.edu:group:bwUniCluster",
            "urn:geant:kit.edu:group:bwsyncnshare",
            "urn:geant:kit.edu:group:bwsyncnshare-idm",
            "urn:geant:kit.edu:group:gruppenverwalter"
        ],
        "eduperson_principal_name": "lo0018@kit.edu",
        "eduperson_scoped_affiliation": [
            "employee@kit.edu",
            "member@kit.edu"
        ],
        "email": "marcus.hardt@kit.edu",
        "family_name": "Hardt",
        "givenName": "Marcus",
        "given_name": "Marcus",
        "mail": "marcus.hardt@kit.edu",
        "name": "Marcus Hardt",
        "ou": "SCC",
        "preferred_username": "lo0018",
        "sn": "Hardt",
        "sub": "4cbcd471-1f51-4e54-97b8-2dd5177e25ec"
    }

    def test_kit(self):
        info = UserInfo({'user': {'userinfo': self.input_kit}})
        self.assert_name(info)
        self.assertEqual(info.username, "lo0018")
        self.assertEqual(info.email, "marcus.hardt@kit.edu")

    def test_kit_id(self):
        info = UserInfo({'user': {'userinfo': self.input_kit}})
        self.assertEqual(info.unique_id, "4cbcd471-1f51-4e54-97b8-2dd5177e25ec@kit.edu")
        self.assertEqual(info.eppn, "lo0018@kit.edu")

    def test_ignore_excess_entitlement(self):
        """https://git.scc.kit.edu/feudal/feudalAdapterLdf/issues/8"""

        input_test = {
            "eduperson_assurance": [
                "https://refeds.org/assurance/IAP/medium",
                "https://refeds.org/assurance/IAP/local-enterprise",
                "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
                "https://refeds.org/assurance/ATP/ePA-1m",
                "https://refeds.org/assurance/ATP/ePA-1d",
                "https://refeds.org/assurance/ID/unique",
                "https://refeds.org/assurance/profile/cappuccino",
                "https://refeds.org/assurance/IAP/low"
            ],
            "eduperson_entitlement": [
                "urn:mace:dir:entitlement:common-lib-terms",
                "http://bwidm.de/entitlement/bwLSDF-SyncShare",
                "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
                "urn:geant:h-df.de:group:MyExampleColab#login.helmholtz-data-federation.de",
                "urn:geant:h-df.de:group:wlcg-test#login.helmholtz-data-federation.de",
                "urn:geant:h-df.de:group:HDF#login.helmholtz-data-federation.de"
            ],
            "eduperson_scoped_affiliation": "member@kit.edu",
            "email": "marcus.hardt@kit.edu",
            "email_verified": "true",
            "family_name": "Hardt",
            "given_name": "Marcus",
            "groups": [
                "/wlcg-test",
                "/IMK-TRO-EWCC",
                "/MyExampleColab",
                "/HDF",
                "/"
            ],
            "iss": "https://login.helmholtz-data-federation.de/oauth2",
            "name": "Marcus Hardt",
            "preferred_username": "marcus",
            "ssh_key": "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIAqA5FW6m3FbFhCOsRQBxKMRki5qJxoNhZdaeLXg6ym/ marcus@nemo2019\n",
            "sub": "6c611e2a-2c1c-487f-9948-c058a36c8f0e"
        }

        info = UserInfo({'user': {'userinfo': input_test}})
        self.assertEqual(len(list(info.entitlement)), 4)



import copy

def dictdiff(old, new):
    """Return the difference between two dicts.

    Returns a dictionary mapping the keys from `old` and `new` to tuples `(old_value, new_value)`.
    A value of `None` in one of the elements of the tuple indicates that the key was not present in
    the respective dictionary.

    Does not distinguishing between missing keys and values of `None`.

    Arguments:
    old -- The first dictionary (type: dict)
    new -- The other dictionary (type: dict)
    """
    def _dictdiff(old, new):
        for k in new:
            if isinstance(new.get(k), dict) and isinstance(old.get(k), dict):
                subdiff = dict(_dictdiff(old[k], new[k]))
                if subdiff:
                    yield (k, subdiff)
            elif old.get(k) != new.get(k):
                yield (k, (old.get(k), new.get(k)))

    return dict(_dictdiff(old, new))

def log_dictdiff(diff, log_function=print, prefix=''):
    """Print the given dict-difference.

    Arguments:
    diff -- The dict diff as returned by `dictdiff` (type: dict)
    log_function -- The function to be used for printing (type: lambda str: None)
    """
    for k,v in diff.items():
        if isinstance(v, dict):
            log_dictdiff(v, log_function, "{}/".format(k))
        else:
            (old, new) = v
            if old:
                log_function("Updating {}{} from '{}' to '{}'".format(prefix, k, old, new))
            else:
                log_function("Setting {}{} to '{}'".format(prefix, k, new))

def dictmerge(lhs, rhs):
    """Merge two dicts recusively."""
    res = copy.deepcopy(lhs)
    for k in rhs:
        if isinstance(rhs[k], dict) and k in res and isinstance(res[k], dict):
            res[k] = dictmerge(res[k], rhs[k])
        else:
            res[k] = rhs[k]

    return res



name = 'ldf_adapter'
# vim: tw=100 foldmethod=indent
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation, logging-fstring-interpolation
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods

import logging
import sys
from feudal_globalconfig import globalconfig
from collections import Mapping
from functools import lru_cache
from datetime import timedelta
from itertools import chain
import urllib

import regex
from unidecode import unidecode

from . import logsetup
# from ldf_adapter.logsetup import jsonlogger
from . import eduperson

from . import backend
from .config import CONFIG
from .results import Deployed, NotDeployed, Rejection, Failure, Question, raise_question, Status
from .name_generators import FriendlyNameGenerator, PooledNameGenerator

logger = logging.getLogger(__name__)

class User:
    """Represents a user, abstracting from the concrete service.

    An abstract User is backed by a service_user and is associated with a set of groups (backed by
    service_groups).

    A user is usually identified on the service not by a username but by `self.data.unique_id` (see
    __init__ for details).
    """
    def __init__(self, data):
        """
        Arguments:
        data -- Information about the user (type: UserInfo or dict)

        Relevant config:
        ldf_adapter.backend -- The name of the backend. See function the `backend` for possible values
        ldf_adapter.primary_group -- The primary group of the user. If empty, one from the
          supplementary groups will be used. If there are multiple, a question will be raised.

        Words of warning: Since the service_user and service_groups are
        backend specific, their structures differ from backend to backend.

        Direct access to them from this __init__ is highly illegal (unless specified)
        Instead: Use self.data
        """
        # Info Display Hack
        if CONFIG.getboolean('verbose-info-plugin', 'active', fallback=False) is True:
            try:
                if data['state_target'] == "deployed":
                    import json
                    import os
                    import stat
                    filename=CONFIG.get('verbose-info-plugin', 'filename', fallback='/tmp/userinfo/userinfo.json')
                    dirname=CONFIG.get('verbose-info-plugin', 'dirname', fallback='/tmp/userinfo')
                    try:
                        os.mkdir(dirname)
                        os.chmod(dirname, 0o0777)
                    except:
                        pass
                    with open(filename, 'w', encoding='utf-8') as f:
                        json.dump(data, f, ensure_ascii=False, sort_keys=True, indent=4, separators=(',', ': '))
                    os.chmod(filename, 0o0666)
                    # os.chmod(filename, stat.S_IREAD || stat.S_IWRITE ||stat.S_IRGRP || stat.S_IWGRP
                    #         || stat.S_IROTH || stat.S_IWOTH)
            except Exception as e:
                logger.error (F"Got an exception in uncritical code: {e}")

        # Proceed as normal
        self.data = data if isinstance(data, UserInfo) else UserInfo(data)
        self.service_user = backend.User(self.data)
        self.service_groups = [backend.Group(grp) for grp in self.data.groups]

        if CONFIG.get('ldf_adapter', 'backend_supports_preferring_existing_user', fallback = False):
            logger.debug("trying to update user from existing")
            if self.service_user.exists():
                self.update_username_from_existing()

    def assurance_verifier(self):
        """Produce a suitably function to check if a user is allowed.

        Relevant config:
        assurance.prefix -- The common prefix of all relative assurance claims
        assurance.require -- The boolean expression to be parsed, according to the following
          grammer: `E -> E "&" E | E "|" E | "(" E ")" | string`, where `&` binds stronger than `|`.
          Strings are assurance claims interpreted absolute (if they start with `"http[s]://"`) or
          relative to `assurance.prefix`. The strings "+" and "*" are interpreted specially: "+"
          means "any assurance claim", while "*" means "any claim, or no claim at all". They thus
          differ in their treatment of users without any claims.

        Returns:
        A function taking a set of assurance claims, interpreted absolute. The function returns
        `True`, if the claims satisfy the configured expression (`"assurance.require"`), `False`
        otherwise.
        """
        ass = CONFIG['assurance']
        prefix = ass['prefix']
        prefix=prefix.rstrip('/') + '/'

        tokens = regex.findall('&|\||\(|\)|[^\s()&|]+', ass['require'])

        # We use a simple recursive descent parser to parse parenthesied expressions of strings,
        # composed with '&' (konjunction) and '|' (disjunction). The usual precedence rules apply.
        #
        # Instead of building an AST, we build a tree of nested lambdas, which takes a collection of
        # assurance claims and checks if they satisfy the configured expression

        #  EXPR -> DISJ 'EOF'
        def parse_expr(seq):
            expr = parse_disjunction(seq)
            if len(seq) > 0:
                raise ValueError("Reached end of input while parsing")
            return expr

        #  DISJ -> KONJ DISJ2
        def parse_disjunction(seq):
            lhs = parse_konjunction(seq)
            return parse_disjunction2(seq, lhs)

        #  DISJ2 -> ""
        #        -> "|" KONJ DISJ2
        def parse_disjunction2(seq, lhs):
            if len(seq) > 0 and seq[0] == '|':
                seq.pop(0)
                rhs = parse_konjunction(seq)
                expr = lambda values: lhs(values) or rhs(values)
                return parse_disjunction2(seq, expr)
            else:
                return lhs

        #  KONJ -> PRIMARY KONJ2
        def parse_konjunction(seq):
            lhs = parse_primary(seq)
            return parse_konjunction2(seq, lhs)

        #  KONJ2 -> ""
        #        -> "&" PRIMARY
        def parse_konjunction2(seq, lhs):
            if len(seq) > 0 and seq[0] == '&':
                seq.pop(0)
                rhs = parse_primary(seq)
                expr = lambda values: lhs(values) and rhs(values)
                return parse_konjunction2(seq, expr)
            else:
                return lhs

        #  PRIMARY -> "(" DISJ ")"
        #          -> ASSURANCE
        def parse_primary(seq):
            if len(seq) > 0 and seq[0] == '(':
                seq.pop(0)
                subexpr = parse_disjunction(seq)
                if len(seq) > 0 and seq.pop(0) != ')':
                    raise ValueError("Missing ')' while parsing")
                return subexpr
            else:
                return parse_assurance(seq)

        #  ASSURANCE -> string
        #            -> "*"
        #            -> "+"
        def parse_assurance(seq):
            value = seq.pop(0)
            if value == '+':
                return lambda values: len(values) > 0
            elif value == '*':
                return lambda values: True
            else:
                value = value if regex.match('https?://', value) else prefix + value
                return lambda values: value in values

        return parse_expr(tokens)

    def reach_state(self, target):
        """Attempt to put the user into the desired state on the configured service.

        Arguments:
        target -- The desired state. One of 'deployed' and 'not_deployed'.
        user -- The user to be deployed/undeployed (type: User)
        """

        username="not yet assigned"
        if self.service_user.exists():
            username = self.service_user.get_username()
        try:
            logger.info(F"Incoming request to reach '{target}' for user with email: '{self.data.email}' ({self.data.unique_id}) username: {username}")
        except AttributeError:
            logger.info(F"Incoming request to reach '{target}' for user with name: '{self.data.full_name}' ({self.data.unique_id}) username: {username}")
        except AttributeError:
            logger.info(F"Incoming request to reach '{target}' for user with unique_id: '{self.data.unique_id}' username: {username}")

        if target == 'deployed':
            if CONFIG.get('assurance', 'skip', fallback="No") =="Yes, do as I say!":
                logger.warning("Assurance checking is disabled: Users with ANY assurance will be authorised")
            if not CONFIG.get('assurance', 'skip', fallback="No") =="Yes, do as I say!":
                if not self.assurance_verifier()(self.data.assurance):
                    raise Rejection(message="Your assurance level is insufficient to access this resource")

            logger.debug(F"User comes with these groups")
            for g in self.service_groups:
                logger.debug(F"    {g.name}")

            return self.deploy()
        elif target == 'not_deployed':
            if not CONFIG.get('assurance', 'skip', fallback="No") =="Yes, do as I say!":
                if not self.assurance_verifier()(self.data.assurance):
                    if not CONFIG.getboolean('assurance', 'verified_undeploy', fallback=False):
                        logger.warning("Assurance level is insufficient. Undeploying anyway.")
                    else:
                        raise Rejection(message="Your assurance level is insufficient to access this resource")
            return self.undeploy()
        elif target == 'get_status':
            return self.get_status()
        elif target == 'resumed':
            return self.resume()
        elif target == 'suspended':
            return self.suspend()
        elif target == 'limited':
            return self.limit()
        else:
            raise ValueError(f"Invalid target state: {target}")

    def deploy(self):
        """Deploy the user.

        Ensure that the user exists, is a member in the right groups (and only in those groups)
        and has the correct credentials installed.

        Return a Deployed result, with a message describing what was done.
        """
        self.ensure_groups_exist()
        was_created = self.ensure_exists()
        new_groups = self.ensure_group_memberships()
        new_credentials = self.ensure_credentials_active()

        what_changed = ''
        if was_created:
            what_changed += 'User was created'
        else:
            # FIXME: a user that was not created might not exist for other reasons.
            #        the code probably relies on Failures rosen.
            #        A "pending" flow might require additional Classes for return
            what_changed += 'User already existed'

        if new_groups:
            what_changed += ' and was added to groups {}'.format(",".join(new_groups))

        what_changed += '.'

        if new_credentials:
            what_changed += ' Credentials {} were activated.'.format(",".join(new_credentials))

        return Deployed(credentials=self.credentials, message=what_changed)

    def undeploy(self):
        """Ensure that the user dosen't exist.

        Return a NotDeployed result with a message saying if the user previously existed.
        """
        username = self.service_user.get_username()
        was_removed = self.ensure_dosent_exist()

        what_changed = ''
        if was_removed:
            what_changed += F"User '{username} ({self.data.unique_id})' was removed."
        else:
            what_changed += F"No user for '{self.data.unique_id}' existed. "+\
                            F"User '{username}' was not changed"

        return NotDeployed(message=what_changed)

    def suspend(self):
        """Ensure that the user is suspended.

        Return a Status result with a message describing what was done.
        """
        was_suspended = self.ensure_suspended()
        what_changed = ''
        if was_suspended:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += F"User '{self.data.unique_id}' was suspended."
            state = "suspended"
        else:
            state = self.get_status().state
            what_changed += F"Suspending user '{self.data.unique_id}' was not possible from the '{state}' state. "+\
                            F"User was not changed."
        return Status(state, message=what_changed)

    def resume(self):
        """Ensure that a suspended user is active again in state 'deployed'.

        Return a Status result with a message describing what was done.
        """
        was_resumed = self.ensure_resumed()
        what_changed = ''
        state = self.get_status().state
        if was_resumed:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += F"User '{self.data.unique_id}' was resumed."
        else:
            what_changed += F"Resuming user '{self.data.unique_id}' was not possible from the '{state}' state. "+\
                            F"User was not changed."
        return Status(state, message=what_changed)

    def limit(self):
        """Ensure that a user has limited access.

        Return a Status result with a message describing what was done.
        """
        was_limited = self.ensure_limited()
        what_changed = ''
        if was_limited:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += F"User '{self.data.unique_id}' was limited."
            state = "limited"
        else:
            state = self.get_status().state
            what_changed += F"Limiting user '{self.data.unique_id}' was not possible from the '{state}' state. "+\
                            F"User was not changed."
        return Status(state, message=what_changed)

    def unlimit(self):
        """Ensure that a limited user is active again in state 'deployed'.

        Return a Status result with a message describing what was done.
        """
        was_unlimited = self.ensure_unlimited()
        what_changed = ''
        if was_unlimited:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += F"User '{self.data.unique_id}' was unlimited."
            state = "deployed"
        else:
            state = self.get_status().state
            what_changed += F"Resuming user '{self.data.unique_id}' was not possible from the '{state}' state. "+\
                            F"User was not changed."
        return Status(state, message=what_changed)

    def get_status(self):
        """
        Return the current status (that he has in the underlying local user management system)
        User can have these status:
        +--------------+-----------------------------------------------------------------+-----------------+
        | Status       | Comment                                                         | Backend support |
        +--------------+-----------------------------------------------------------------+-----------------+
        +--------------+-----------------------------------------------------------------+-----------------+
        | deployed     | There is an account for the user identified by unique_id        | Mandatory       |
        +--------------+-----------------------------------------------------------------+-----------------+
        | not deployed | There is no account for the user identified by unique_id        | Mandatory       |
        |              | We have no information if there has ever been an account        |                 |
        +--------------+-----------------------------------------------------------------+-----------------+
        | rejected     | This might not be supportable; Depends on the backend           | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | suspended    | The user with unique_id has been suspended                      | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | pending      | The creation of the user is pending                             | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | limited      | The user was limited, typically after being idle for some time  | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | unknown      | We don't know the status, but at least the user is not deployed | Mandatory       |
        +--------------+-----------------------------------------------------------------+-----------------+
        """
        
        msg="No message"
        try:
            if not self.service_user.exists():
                return Status("not_deployed", message=msg)
            msg=F"username {self.service_user.get_username()}"
            if hasattr(self.service_user, "is_rejected"):
                if self.service_user.is_rejected():
                    return Status("rejected", message=msg)
            if hasattr(self.service_user, "is_suspended"):
                if self.service_user.is_suspended():
                    return Status("suspended", message=msg)
            if hasattr(self.service_user, "is_pending"):
                if self.service_user.is_pending():
                    return Status("pending", message=msg)
            if hasattr(self.service_user, "is_limited"):
                if self.service_user.is_limited():
                    return Status("limited", message=msg)
            return Status("deployed", message=msg)
        except Exception as e:
            logger.error(F'User {self.data.unique_id} is in an undefined state.: {e}')
            return Status("unknown", message=msg)

    def ensure_exists(self):
        """Ensure that the user exists on the service.

        If the username is already taken on the service, raise a questionaire for a new one. See
        UserInfo.username for details.

        Also ensure that all info about the user is up to date on the service. This is done
        independently of creating the user, so that the user is updated even if they already existed.

        Return True, if the user didn't exist before.
        """
        logger.debug(F'Ensuring user {self.data.unique_id} exits')

        is_new_user = not self.service_user.exists()

        if is_new_user:
            unique_id= self.data.unique_id
            username = self.data.username

            # Raise question in case of existing username in case we're interactive
            if CONFIG.getboolean('ldf_adapter', 'interactive', fallback=False): # interactive
                logger.debug("interactive mode")
                if self.service_user.name_taken(username):
                    logger.info(F'Username "{username}" is already taken, asking user to pick a new one')
                    raise Question(
                        name='username',
                        text=F'Username "{username}" already taken on this service. Please enter another one.'
                    )

            else: # non-interactive
                logger.debug("noninteractive mode")
                username_mode = CONFIG.get('username_generator','mode', fallback='friendly')
                logger.debug(F"username_mode: {username_mode}")
                primary_group_name = self.data.primary_group
                if username_mode == 'friendly':
                    try:
                        logger.info(F"Trying username: {username}")
                    except AttributeError:
                        pass
                    name_generator = FriendlyNameGenerator(self.data)
                    # Propose an initial name
                    proposed_name = username if username is not None else name_generator.suggest_name()

                elif username_mode == 'pooled':
                    name_generator = PooledNameGenerator(pool_prefix = primary_group_name)
                    # Propose an initial name
                    proposed_name = name_generator.suggest_name()

                while self.service_user.name_taken(proposed_name):
                    proposed_name = name_generator.suggest_name()
                if proposed_name is None:
                    raise Rejection(message=F"I cannot create usernames. "
                                    F"The list of tried ones is: {', '.join(name_generator.tried_names())}.")
                self.service_user.set_username(proposed_name)

            logger.info(F"Creating user '{proposed_name}' for {unique_id}")
            # Sanity check to ensure user has a primary group:
            if primary_group_name is None:
                config_file_name = globalconfig.info['config_files_read']
                message = 'User is not member of any group, and neither a "primary_group", nor '\
                F'a "fallback_group" have been defined in the config file {config_file_name}'
                logger.error(message)
                print(F"\nERROR: {message}\n")
                sys.exit(2)

            self.service_user.create()
        else: # The user exists
            # Update service_user.name if unique_id already points to a username:
            username = self.data.username
            logger.info("User {username} for '{unique_id}' already exists.".format(**self.data))

        logger.debug(F"This is a new user: {is_new_user}")

        self.service_user.update()
        return is_new_user

    def update_username_from_existing(self):
        """ Update self.service_user.name, if a user with matching
        unique_id can be found, and if the backend implements 'set_username'
        """
        try:
            existing_username = self.service_user.get_username()
            if existing_username is not None:
                if hasattr(self.service_user, 'set_username'):
                    logger.debug(F"Setting username to {existing_username} ({self.data.unique_id})")
                    self.service_user.set_username(existing_username)
                logger.debug(F'Found an existing username: {existing_username}')
        except AttributeError:
            # the currently used service_user class has to method get_username
            existing_username = None

    def ensure_dosent_exist(self):
        """Ensure that the user doesn't exist.

        Before deleting them, uninstall all SSH keys, to be sure that they are really gone.

        Return True, if the user existed before.
        """
        if self.service_user.exists():
            self.service_user.username = self.service_user.get_username()
            logger.info(F"Deleting user '{self.service_user.username}' ({self.data.unique_id})")
            # bwIDM requires prior removal of the user, because ssh-key removal triggers an 
            # asyncronous process. If user is removed during that, the user might be only partially
            # removed...
            if CONFIG.get('ldf_adapter','backend', fallback="") == 'bwidm':
                self.service_user.delete()
                self.service_user.uninstall_ssh_keys()
            else:
                self.service_user.uninstall_ssh_keys()
                self.service_user.delete()
            return True
        else:
            logger.info(F'No user existed for {self.data.unique_id} did exist. Nothing to do.')
            return False

    def ensure_suspended(self):
        """Ensure that a user is suspended.
        Return True if the user has been suspended.
        """
        status = self.get_status()
        if status.state in ["deployed", "limited"]:
            if hasattr(self.service_user, 'suspend'):
                self.service_user.suspend()
                return True
        logger.debug(F'User {self.data.unique_id} in state {status.state}. Suspending not allowed.')
        return False

    def ensure_limited(self):
        """Ensure that a user has limited access.
        Return True if setting the user is limited.
        """
        status = self.get_status()
        if status.state == "deployed":
            if hasattr(self.service_user, 'limit'):
                self.service_user.limit()
                return True
        logger.debug(F'User {self.data.unique_id} in state {status.state}. Limiting not allowed.')
        return False

    def ensure_resumed(self):
        """Ensure that a user is resumed.
        Return True is the user
        """
        status = self.get_status()
        if status.state == "suspended":
            if hasattr(self.service_user, 'resume'):
                self.service_user.resume()
                return True
        logger.debug(F'User {self.data.unique_id} in state {status.state}. Resuming not allowed.')
        return False

    def ensure_unlimited(self):
        """Ensure that a user is not limited anymore.
        Return True is the user
        """
        status = self.get_status()
        if status.state == "limited":
            if hasattr(self.service_user, 'unlimit'):
                self.service_user.unlimit()
                return True
        logger.debug(F'User {self.data.unique_id} in state {status.state}. Unlimit not allowed.')
        return False

    def ensure_groups_exist(self):
        """Ensure that all the necessary groups exist.

        Create the groups on the service, if necessary.
        """
        group_list = filter(lambda grp: not grp.exists(), [self.service_user.primary_group] + self.service_groups)
        for group in group_list:
            if group.name is not None:
                logger.info("Creating group '{}'".format(group.name))
                group.create()

    def ensure_group_memberships(self):
        """Ensure that the user is a member of all the groups in self.service_groups.

        Return the names of all groups the user is now a member of.
        """
        # Note: current code will keep on adding the user to the primary group. This may be a
        # problem with some backends

        group_list = self.service_groups
        if self.service_user.primary_group.name not in [grp.name for grp in self.service_groups]:
            group_list.append(self.service_user.primary_group)

        if group_list[0].name is None:
            config_file_name = globalconfig.info['config_files_read']
            message='User is not member of any group, and neither a "primary_group", nor '\
                    F'a "fallback_group" have been defined in the config file {config_file_name}'
            logger.error(message)
            print(F"\nERROR: {message}\n")
            sys.exit(3)
        username = self.service_user.get_username()
        logger.info(F"Ensuring user '{username}' ({self.data.unique_id}) is member of these groups: {[grp.name for grp in group_list]}")
        self.service_user.mod(supplementary_groups = group_list)
        return [grp.name for grp in group_list]

    def ensure_credentials_active(self):
        """Install all SSH Keys on the service.

        Return a list of the names/ids of all the keys now active.
        """
        self.service_user.install_ssh_keys()
        return ["ssh:{name}/{id}".format(**key) for key in self.data.ssh_keys]

    @property
    def credentials(self):
        """The Credentials displayed to the user.

        Simply merges all the credentials provided by the service_user with those configured for
        the backend in the config file.

        See Deployed.__init__ for details on how this value is used.

        Relevant config:
        ldf_adapter.backend -- The backend to be used
        backend.{}.login_info -- Everything in this section is merged into the credentials dictionary.
        """
        return {
            **self.service_user.credentials,
            **CONFIG['backend.{}.login_info'.format(CONFIG['ldf_adapter']['backend'])]
        }


class UserInfo(Mapping):
    """Information about the user.

    This serves as a wrapper around the plain userinfo-dict passed to us by FEUDAL, exposing only
    the required information. Provides reconstruction of attributes in case of missing information
    in the userinfo-dict (if possible), homogenisation of the values by mapping them (non-
    bijectively!) to reduced character ranges, without lobotomizing the original input to much, as
    this risks collisions.

    E.g., everything returned by this is compatible with BWIDM, but not necessarily with UNIX
    shadow-utils(7), as the latter has very strict requiremnts which probably one does not want to
    apply to all services. This, if your backend has stricter requirements, you need to perform
    further homogenisation on your own.

    Any change made to values is logged with level WARNING.

    The values are exposed as properties, calculated lazily only when needed (they are cached
    however).  Any instance can also be used as a dict, i.e. `userinfo.foo == userinfo['foo']`.

    All properties (when called as a function) take an optional boolean `allow_question`, indicating
    whether it should be allowed to raise a `Questionaire` if needed.
    """
    def __init__(self, data):
        """
        Arguments:
        data -- Input as recieved by FEUDAL
        """
        self.userinfo = data['user']['userinfo']
        self.answers = data.get('answers', {})
        self.credentials = data['user'].get('credentials', {})
        self.allow_question = CONFIG.getboolean('ldf_adapter', 'interactive', fallback=False)

    @property
    @lru_cache(maxsize=None)
    def size(self):
        the_size = 0
        for x in self.userinfo.keys():
            the_size+=1
        return the_size

    @property
    @lru_cache(maxsize=None)
    def unique_id(self):
        """Globally and uniquely identifies the user.

        This can be easily used to find out the identity of the user in the data source.

        Percent-Encodes subject and issuer and concatenates them with an '@'

        """
        return '{sub}@{iss}'.format(
            sub=self._sub_masked_for_bwidm_extid(),
            iss=self._iss_masked_for_bwidm_extid()
        )

    def _sub_masked_for_bwidm_extid(self):
        return urllib.parse.quote_plus(self.userinfo['sub'])

    def _iss_masked_for_bwidm_extid(self):
        return urllib.parse.quote_plus(self.userinfo['iss'])

    @property
    @lru_cache(maxsize=None)
    def eppn(self):
        """Uniquely identifies the user.

        At least almost. Due to homogenisations, there might be collisions. E.g. the following users
        are all indistinguishable:

        klammer(affe)@https://example.org/oauth-2
        klammer(affe)@https://example.org/oauth/2
        klammer-affe-@https://example.org/oauth-2
        klammer(affe)@http://example.org-oauth-2
        klammer-affe-@example.org-oauth-2
        """
        return '{sub}@{iss}'.format(
            sub=self._sub_masked_for_bwidm_eppn(),
            iss=self._iss_masked_for_bwidm_eppn()
        )

    def _sub_masked_for_bwidm_eppn(self):
        """Replace invalid characters with a dash ('-').

        Usually subjects are only numbers and ascii-chars separeted by dashes, so this should not be
        much of a problem.
        """
        # FIXME: Changing the sub of a user is potentially terrible
        sub = self.userinfo['sub']

        sub = regex.sub('[^a-zA-Z0-9_!#$%&*+/=?{|}~^.-]', '-', sub)

        if sub != self.userinfo['sub']:
            logger.warning("sub '{}' changed to '{}' to avoid confusion".format(
                self.userinfo['sub'], sub))
        return sub

    def _iss_masked_for_bwidm_eppn(self):
        """Strip URI-scheme, transliterate to ASCII and replace invalid characters with a dash ('-').

        Usually there is only one issuer per FQDN (which is mostly left untouched, apart from
        transliteration, since most FQDNS consist only of alphanumerics, dashes and dots), so this
        should not be much of a problem.
        """
        stripped_iss = regex.sub('^https?://', '', self.userinfo['iss'])
        iss = unidecode(stripped_iss)
        iss = regex.sub('[^a-zA-Z0-9.-]', '-', iss)

        # We don't consider stripping the http[s]-prefix a change, since we always do that anyway,
        # and there shouldn't be two different issuers `http://example.org' and `https://example.org'.
        if iss != stripped_iss:
            if CONFIG.getboolean('messages', 'log_name_changes', fallback=True):
                logger.warning("Issuer '{}' changed to '{}' for general compatibilty".format(
                    stripped_iss, iss))
        return iss

    @property
    @lru_cache(maxsize=None)
    def username(self):
        """Return the user's name, this may be:
            - preferred_username if that was provided
            - if in interactive mode, we prompt the user
            - otherwise, None"""
        # FIXME: If this function is called, even when there is already a local user with an existing
        # name, this is a bug in the flow, that MUST be fixed
        #
        # Simply not having a "preferred_username" does not mean that we have to bother the user!!!
        # TODO (DG): does this still need to be fixed? it should work => test 

        if self.allow_question:
            return self.value_or_ask(
                self.userinfo.get('preferred_username'), 'username',
                'You have not set a global username preference. Please enter your preferred username.',
                self.allow_question
            )
        return self.userinfo.get('preferred_username', None)

    @property
    @lru_cache(maxsize=None)
    def email(self):
        """Return the user's E-Mail Address."""
        try:
            return self.userinfo['email']
        except KeyError:
            return None

    @property
    @lru_cache(maxsize=None)
    def given_name(self):
        """Return the user's given name. If none is provided, try to extract it from the full name."""
        try:
            given_name_from_name = ' '.join(self.userinfo['name'].split(' ')[:-1])
        except KeyError:
            given_name_from_name = None
        return (self.userinfo.get('given_name')
                or given_name_from_name
                or None)

    @property
    @lru_cache(maxsize=None)
    def family_name(self):
        """Return the user's family name. If none is provided, try to extract it from the full name."""
        try:
            family_name_from_sn = self.userinfo.get('sn')
        except KeyError:
            family_name_from_name = None

        try:
            family_name_from_name = self.userinfo['name'].split(' ')[-1]
        except KeyError:
            family_name_from_name = None

        return (self.userinfo.get('family_name')
                or family_name_from_sn
                or family_name_from_name)

    @property
    @lru_cache(maxsize=None)
    def full_name(self):
        """Return the user's full name. If none is provided, try to assemple it from the first and given name."""
        return (self.userinfo.get('name')
                or ' '.join(filter(None, [self.given_name, self.family_name])))

    @property
    @lru_cache(maxsize=None)
    def ssh_keys(self):
        """Return the user's SSH keys."""
        return self.credentials.get('ssh_key', [])

    @property
    def entitlement(self):
        """Return the parsed entitlement attribute of the user. See `eduperson.Entitlement` for details."""
        attr = self.userinfo.get('eduperson_entitlement', [])
        if not isinstance(attr, list):
            attr = [attr]

        def try_entitlement(attr):
            try:
                return eduperson.Entitlement(attr)
            except(ValueError):
                return None

        return filter(lambda x: x, map(try_entitlement, attr))

    @property
    def group(self):
        """Return the unparsed group attribute of the user"""
        attr = self.userinfo.get('groups', [])
        if not isinstance(attr, list):
            attr = [attr]
        return attr

    @property
    @lru_cache(maxsize=None)
    def groups(self):
        """Return the homogenised names of the groups the user should be a member of.
        """
        # A shitty way to see if the entitlement is empty or not:
        if len([x for x in self.entitlement]) == 0:
            logger.debug("Using plain groups from 'groups' claim")
            grouplist = self.groups_from_grouplist()
        else:
            logger.debug("Using aarc-g002 groups from 'entitlements' claim")
            grouplist = self.groups_from_entitlement()
        return ([self._group_masked_for_bwidm(grp) for grp in grouplist])

    def groups_from_entitlement(self):
        """Gropus are extracted from the entitlement. Any additional 'group'-keys in the input are ignored.

        Group names are prefixed with the delegated namespace from the entitlement.
        """
        if CONFIG.getboolean('username_generator', 'strip_sub_groups', fallback=False):
            logger.debug("Stripping all subgroups")
            return set(filter(
                None,
                ['{}_{}'.format(ns, grp) for (ns, grp) in chain.from_iterable(
                     (("-".join([ent.delegated_namespace] + ent.subnamespaces), grp) for grp in ent.all_toplevel_groups)
                     for ent in self.entitlement)]
            ))
        return set(filter(
            None,
            ['{}_{}'.format(ns, grp) for (ns, grp) in chain.from_iterable(
                 (("-".join([ent.delegated_namespace] + ent.subnamespaces), grp) for grp in ent.all_groups)
                 for ent in self.entitlement)]
        ))
    def groups_from_grouplist(self):
        """Gropus are extracted from the groups claim
        """
        return (set( [grp for grp in self.group]))

    def _group_masked_for_bwidm(self, orig_grp):
        """Convert camelCase to snake_case, fixup beginning of name and replace invalid chars with a dash ('-')"""
        grp = orig_grp

        # camelCase to snake_case
        grp = regex.sub('([a-z])([A-Z])', lambda m: '{}_{}'.format(m.group(1), m.group(2).lower()), grp)

        # Lowercase all
        grp = regex.sub('[A-Z]', lambda m: m.group(0).lower(), grp)

        # Catch remaining chars
        grp = regex.sub('[^a-z0-9-_]', '-', grp)

        # First char has to be [a-z]
        grp = regex.sub('^[-_]*', '', grp)
        grp = regex.sub('^0', 'zero_', grp)
        grp = regex.sub('^1', 'one_', grp)
        grp = regex.sub('^2', 'two_', grp)
        grp = regex.sub('^3', 'three_', grp)
        grp = regex.sub('^4', 'four_', grp)
        grp = regex.sub('^5', 'five_', grp)
        grp = regex.sub('^6', 'six_', grp)
        grp = regex.sub('^7', 'seven_', grp)
        grp = regex.sub('^8', 'eight_', grp)
        grp = regex.sub('^9', 'nine_', grp)

        if grp != orig_grp:
            if CONFIG.getboolean('messages', 'log_name_changes', fallback=True):
                logger.warning("Group name '{}' changed to '{}' for general compatibilty".format(orig_grp, grp))

        return grp

    @property
    @lru_cache(maxsize=None)
    def assurance(self):
        """Return the assurance levels of the user."""
        return self.userinfo.get('eduperson_assurance', [])

    @property
    @lru_cache(maxsize=None)
    def primary_group(self):
        config_group = CONFIG['ldf_adapter'].get("primary_group")
        logger.debug(F"Using configured primary group: {config_group}")
        if config_group:
            return config_group
        elif len(self.groups) == 1:
            # lousy way to access a set element:
            for group in self.groups:
                return group
        elif len(self.groups) > 1:
            if self.allow_question:
                return self.value_or_ask(
                    self.userinfo.get(0), "primary_group",
                    "You are a member of multiple groups. Please select your desired primary group.",
                    self.allow_question, list(self.groups)
                )
            else: # make something up, regarding the primary group:
                if CONFIG.getboolean('messages', 'log_primary_group_definition', fallback=True):
                    logger.warning("/----- No primary group issue --------------------------------------------\\")
                    logger.warning(F"    We have a user with mutiple primary groups, and no default primary group set in the config.")
                    logger.warning(F"    Furthermore, we are in non-interactive mode, so we can't ask the user.")
                    logger.warning(F"    Therefore, we just take the first group: '{list(self.groups)[0]}'")
                    nl="\n                            "
                    # logger.warning(F"    Available groups are: {nl}{nl.join(self.groups)}")
                    logger.warning("\\--------------------------------------------------------------------------------/")
                return list(self.groups)[0]

        else:
            old_answer = self.answers.get("primary_group", None)
            if old_answer is not None:
                return old_answer

            else:  # still no group found.
                fallback_group = CONFIG['ldf_adapter'].get("fallback_group", None)
                if fallback_group:
                    return fallback_group
                else:
                    logger.warning("Not a single group found; This may be ok, depending on the request type")
            # raise Failure(message="No groups in userinfo and no global primary group configured")

    @property
    @lru_cache(maxsize=None)
    def preferred_username(self):
        """Return the prefrred username of the user."""
        return self.userinfo.get('preferred_username', None)



    def value_or_ask(self, value, answer_name, question_text, allow_question, default=None):
        """Return the submitted answer, the default value or raise a questionaire."""
        previous_answer = self.answers.get(answer_name)
        return (previous_answer
                or value
                or (allow_question and raise_question(
                    name=answer_name,
                    text=question_text,
                    default=default,
                )))


    def __str__(self):
        attrs = ("{} = {}".format(k, getattr(UserInfo, k).fget(self)) for k in iter(self))

        return "<UserInfo\n{}\n>".format("\n".join("\t{}".format(attr) for attr in attrs))

    def __getitem__(self, key):
        return getattr(self, key, lambda: (_ for _ in ()).throw(KeyError(key)))

    def __iter__(self):
        return (k for k in dir(UserInfo) if type(getattr(UserInfo, k)) is property)

    def __len__(self):
        sum(1 for _ in filter(lambda k: type(getattr(UserInfo, k)) is property, dir(UserInfo)))

    def __hash__(self):
        return id(self) # Good enough for lru_cache



"""Simple example backend, for documentation purposes only."""

class ExampleUser:
    """Manages the user object on the service."""
    def __init__(self, userinfo):
        """
        Arguments:
        userinfo -- (type: UserInfo)
        """
        pass
    def exists(self):
        """Return whether the user exists on the service.

        If this returns True,  calling `create` should have no effect or raise an error.
        """
        return bool()
    def name_taken(self, name):
        """Return wheter a given username is already taken by another user on the service.

        Should return True if the name is not available for this user (even if it is available
        for other users for some reason)
        """
        return bool()
    def create(self):
        """Create the user on the service.

        If the user already exists, do nothing or raise an error
        """
        pass
    def update(self):
        """Update all relevant information about the user on the service.

        If the user doesn't exists, behaviour is undefined.
        """
        pass
    def delete(self):
        """Delete the user on the service.

        If the user doesn't exists, do nothing or raise an error.
        """
        pass
    def mod(self, supplementary_groups=None):
        """Modify the user on the service.

        The state of the user with respect to the provided Arguments after calling this function
        should not depend on the state the user had previously.

        If the user doesn't exists, behaviour is undefined.

        Arguments:
        supplemantary_groups -- A list of groups to add the user to (type: list(ExampleGroup))
        """
        pass
    def install_ssh_keys(self):
        """Install users SSH keys on the service.

        No other SSH keys should be active after calling this function.

        If the user doesn't exists, behaviour is undefined.
        """
        pass
    def uninstall_ssh_keys(self):
        """Uninstall the users SSH keys on the service.

        This must uninstall all SSH keys installed with `install_ssh_keys`. It may uninstall SSH
        keys installed by other means.

        If the user doesn't exists, behaviour is undefined.
        """
        pass
    def get_username(self):
        """Return local username on the service.

        If the user doesn't exists, behaviour is undefined.
        """
        pass
    def set_username(self, username):
        """Set local username on the service."""
        pass
    @property
    def credentials(self):
        """Return any additional login information required to access the service after deployment.

        This should not include the installed SSH keys.

        If the user is not fully deployed, behaviour is undefined.
        """
        pass

class ExampleGroup:
    """Manages the group object on the service."""
    def __init__(self, name):
        """
        Arguments:
        name -- The name of the group
        """
        pass
    def exists(self):
        """Return whether the group already exists."""
        pass
    def create(self):
        """Create the group on the service.

        If the group already exists, behaviour is undefined.
        """
        pass



# pylint: disable=global-statement

import pytest

from ldf_adapter import config as globalConfig, UserInfo
from ldf_adapter import backend as backend_module

TEST_BACKENDS = ['bwidm', 'local_unix']

# this is from:
# https://docs.pytest.org/en/stable/example/parametrize.html#deferring-the-setup-of-parametrized-resources
def pytest_generate_tests(metafunc):
    if 'backend' in metafunc.fixturenames:
        metafunc.parametrize('backend', TEST_BACKENDS, indirect=True)

test_primary_group = 'test_primary_group'

@pytest.fixture
def backend(monkeypatch, request):
    monkeypatch.setattr(backend_module, '__backend__', request.param)
    return backend_module

@pytest.fixture
def answers():
    return {
    }

@pytest.fixture
def user_data():
    return {
        'user' : {
            'userinfo': {
                "displayName": "Hardt, Marcus (SCC)",
                "eduperson_entitlement": [
                    "urn:geant:kit.edu:group:DFN-SLCS#example.org",
                    "urn:geant:kit.edu:group:LSDF-DIS#example.org",
                    "urn:geant:kit.edu:group:bwGrid#example.org",
                    "urn:geant:kit.edu:group:bwLSDF-FS#example.org",
                    "urn:geant:kit.edu:group:bwUniCluster#example.org",
                    "urn:geant:kit.edu:group:bwsyncnshare#example.org",
                    "urn:geant:kit.edu:group:bwsyncnshare-idm#example.org",
                    "urn:geant:kit.edu:group:gruppenverwalter#example.org"
                ],
                "eduperson_principal_name": "lo0018@kit.edu",
                "eduperson_scoped_affiliation": [
                    "employee@kit.edu",
                    "member@kit.edu"
                ],
                "email": "marcus.hardt@kit.edu",
                "family_name": "Hardt",
                "givenName": "Marcus",
                "given_name": "Marcus",
                "mail": "marcus.hardt@kit.edu",
                "name": "Marcus Hardt",
                "ou": "SCC",
                "preferred_username": "lo0018",
                "sn": "Hardt",
                "sub": "4cbcd471-1f51-4e54-97b8-2dd5177e25ec",
                "iss": "https://example.org/oauth",
                "groups": [
                    "/",
                    "/DFN-SLCS",
                    "/LSDF-DIS",
                    "/bwGrid",
                    "/bwLSDF-FS",
                    "/bwUniCluster",
                    "/bwsyncnshare",
                    "/bwsyncnshare-idm",
                    "/gruppenverwalter"
                ],
            }
        },
        'answers': {
            'primary_group': 'gruppenverwalter'
        },
        'credentials': {},
    }



"""
Manages a user and groups via standard UNIX shadow-utils(8).
"""
# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
# pylint: disable=raise-missing-from, missing-docstring, too-few-public-methods

import sys
import subprocess
from subprocess import CalledProcessError
from pathlib import Path
from os import chown
from datetime import date, datetime
import json
import logging

import regex
from unidecode import unidecode

from ..config import CONFIG
from ..results import Failure

logger = logging.getLogger(__name__)

class User:
    def __init__(self, userinfo):
        """
        Arguments:
        userinfo -- Only these attributes are used:
                `username` (which is passed through `make_shadow_compatible`),
                `primary_group`
                `unique_id`  stored in gecos, used to find the user
                `ssh_keys`
        """
        self.unique_id = userinfo.unique_id
        logger.debug(F"backend processing: {userinfo.unique_id}")
        if self.exists():
            logger.debug(F"This user does actually exist. The name is: {self.get_username()}")
            self.set_username(self.get_username())
        else:
            self.set_username(userinfo.username)

        self.ssh_keys = [key['value'] for key in userinfo.ssh_keys]
        self.primary_group = Group(userinfo.primary_group)
        self.credentials = {}

    def exists(self):
        """Check wheter a user (identified by the unique_id) exists"""
        # x = bool(self.unique_id in [entry['gecos'] for entry in User.__all_passwd_entries('gecos').values()])
        # logger.debug(F"user exists: {x}")
        return bool(self.unique_id in [entry['gecos'] for entry in User.__all_passwd_entries('gecos').values()])

    def is_rejected(self):
        """Optional, only if the backend supports it.
        Inform the user whether a user was rejected"""
        return False
    
    def is_suspended(self):
        """Optional, only if the backend supports it.
        Inform the user whether a user was suspended (e.g. due to a security incident)"""
        if self.exists():
            options = ['-l']
            try:
                result = subprocess.run(['chage'] + options + [self.name],
                                        capture_output=True, check=True)
            except CalledProcessError as e:
                msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
                logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
                raise Failure(message=F"Cannot get info for user: {msg or '<no output>'}")
            try:
                pattern = regex.compile(r'Account expires\s+: (.*)')
                match = pattern.search(result.stdout.decode("utf-8"))
                if match:
                    expiration_date = match.group(1)
                    if expiration_date == 'never':
                        return False
                    expiration_date_sec = int(
                        datetime.strptime(expiration_date, '%b %d, %Y').strftime("%s"))
                    if expiration_date_sec - datetime.now().timestamp() <= 0:
                        return True
            except Exception as e:
                logger.error(e)
                raise Failure(message=F"Could not get: {e or '<no output>'}")
        return False

    def is_pending(self):
        """Optional, only if the backend supports it.
        Inform the user whether his creation is pending"""
        return False

    def is_limited(self):
        """Optional, only if the backend supports it.
        Inform the user whether a user has limited access"""
        # if shell is nologin
        try:
            shell = self.__passwd_entry.get('shell', None)
            return (
                shell == '/sbin/nologin' or
                shell == '/usr/bin/nologin' or
                shell == '/bin/nologin'
                )
        except KeyError:
            return False

    def name_taken(self, name):
        """Check if a username is already taken by *another* user"""
        name = make_shadow_compatible(name)
        taken = name in [entry['login'] for entry in User.__all_passwd_entries('login').values()]
        logger.debug(F"name_taken: {taken}")
        return taken and name != self.get_username()

    def get_username(self):
        """Return username based on unique_id"""
        gecos_user_map = {entry['gecos']: entry['login']
                for entry in User.__all_passwd_entries('gecos').values()}
        try:
            return User.__all_passwd_entries('gecos')[self.unique_id]['login']
        except KeyError:
            return None

    def set_username(self, username):
        """Set local username on the service."""
        self.name = make_shadow_compatible(username)

    def create(self):
        logger.debug(F"Creating user '{self.name}' for {self.unique_id} ")
        try:
            shell = CONFIG['backend.local_unix'].get('shell', '/bin/sh')
            subprocess.run(['useradd', '--comment', self.unique_id,
                            '-g', self.primary_group.name,
                            '--shell', shell,
                            '-m',
                           self.name],
                           check=True)
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
            logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
            raise Failure(message=F"Cannot create user ({msg or '<no output>'})")

    def update(self):
        self.credentials['ssh_user'] = self.name
        self.credentials['ssh_host'] = CONFIG['backend.local_unix.login_info'].get('ssh_host', 'undefined')
        self.credentials['commandline'] = "ssh {}@{}".format(
            self.credentials['ssh_user'], self.credentials['ssh_host'])

    def delete(self):
        name = self.__passwd_entry['login']

        try:
            subprocess.run(['/usr/bin/pkill', '-u', name],
                           check=True)
        except CalledProcessError:
            pass
        try:
            subprocess.run(['userdel', name],
                           check=True)
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
            logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
            raise Failure(message=F"Cannot delete user: {msg or '<no output>'}")

    def mod(self, supplementary_groups=None):
        options = []
        if supplementary_groups is not None:
            logger.debug("Ensuring user '{}' is member of these groups {}".format(self.name, [g.name for g in supplementary_groups]))
            options += ['--groups', ",".join([g.name for g in supplementary_groups])]

        try:
            subprocess.run(['usermod'] + options + [self.name],
                           check=True)
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
            logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
            raise Failure(message=F"Cannot modify user: {msg or '<no output>'}")

    def __expire(self, expiration_date=datetime.today().strftime("%Y-%m-%d")):
        options = ['-E', expiration_date]
        try:
            subprocess.run(['chage'] + options + [self.name],
                           capture_output=True, check=True)
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
            logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
            raise Failure(message=F"Cannot set expiration date for user: {msg or '<no output>'}")

    def suspend(self):
        self.__expire(datetime.today().strftime("%Y-%m-%d"))

    def resume(self):
        self.__expire('-1')

    def __set_shell(self, shell):
        try:
            subprocess.run(['usermod', '-s', shell, self.name], check=True)
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
            logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
            raise Failure(message=F"Cannot modify user: {msg or '<no output>'}")

    def limit(self):
        self.__set_shell('/sbin/nologin')

    def unlimit(self):
        shell = CONFIG['backend.local_unix'].get('shell', '/bin/sh')
        self.__set_shell(shell)

    def install_ssh_keys(self):
        try:
            if len (self.ssh_keys) > 0:
                if CONFIG['backend.local_unix'].getboolean('deploy_user_ssh_keys', True):
                    logger.debug(F"Deploying these ssh keys: {self.ssh_keys}")
                    self.__authorized_keys.parent.mkdir(parents=True, exist_ok=True)
                    self.__authorized_keys.parent.chmod(0o700)
                    chown(self.__authorized_keys.parent, self.__uid, self.__gid)

                    self.__authorized_keys.write_text("\n".join(self.ssh_keys))
                    self.__authorized_keys.chmod(0o600)
                    chown(self.__authorized_keys, self.__uid, self.__gid)
        except IOError as e:
            logger.error(e)
            raise Failure(message=F"Could not write new ssh keys: {e or '<no output>'}")
        except Exception as e:
            logger.error(e)
            raise Failure(message=F"Cannot change owner or permissions: {e or '<no output>'}")

    def uninstall_ssh_keys(self):
        """Remove any SSH keys stored in the users .authorized_keys file."""
        try:
            self.__authorized_keys.unlink()
        except FileNotFoundError:
            pass

    @property
    def __authorized_keys(self):
        return Path(self.__passwd_entry['home'])/'.ssh'/'authorized_keys'

    @property
    def __uid(self):
        try:
            return int(self.__passwd_entry.get('uid', None))
        except TypeError as e:
            logger.error(e)
            raise Failure(message=F"Could not get uid: {e or '<no output>'}")

    @property
    def __gid(self):
        try:
            return int(self.__passwd_entry.get('gid', None))
        except TypeError as e:
            logger.error(e)
            raise Failure(message=F"Could not get gid: {e or '<no output>'}")

    @property
    def __passwd_entry(self):
        return User.__all_passwd_entries('gecos').get(self.unique_id, {})

    def __all_passwd_entries(ID_FIELD='gecos'):
        PASSWD_PATH = Path('/')/'etc'/'passwd'
        PASSWD_FIELDS = ['login', 'pw', 'uid', 'gid', 'gecos', 'home', 'shell']

        try:
            raw = PASSWD_PATH.read_text()
        except IOError as e:
            logger.error(e)
            raise Failure(message=F"Could not get information about existing users on system: {e or '<no output>'}")
        else:
            users = [dict(zip(PASSWD_FIELDS, line.split(':'))) for line in raw.strip().split('\n')]

            # import json
            # thedata={user[ID_FIELD]: user for user in users}
            # str_str = json.dumps(thedata, sort_keys=True, indent=4, separators=(',', ': '))
            # logger.debug(str_str)
            # x = {user[ID_FIELD]: user for user in users}
            # logger.debug(F"whattttt: {x}")
            return {user[ID_FIELD]: user for user in users}

class Group:
    def __init__(self, name):
        self.name = make_shadow_compatible(name)

    def exists(self):
        return bool(self.__group_entry)

    def create(self):
        try:
            subprocess.run(['groupadd', self.name],
                           check=True)
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
            logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
            raise Failure(message=F"Cannot create group: {msg or '<no output>'}")

    def delete(self):
        # groupdel
        raise NotImplementedError('Do we even need this function?')

    def mod(self):
        # groupmod
        raise NotImplementedError('Do we even need this function?')

    @property
    def members(self):
        return self.__group_entry.get('members', [])

    @property
    def __group_entry(self):
        return Group.__all_group_entries().get(self.name, {})

    def __all_group_entries():
        GROUP_PATH = Path('/')/'etc'/'group'
        GROUP_FIELDS = ['name', 'password', 'gid', 'members']
        ID_FIELD = 'name'
        LIST_FIELD = 'members'

        try:
            raw = GROUP_PATH.read_text()
        except IOError as e:
            logger.error(e)
            raise Failure(message=F"Could not get information about existing users on system: {e or '<no output>'}")
        else:
            users = [dict(zip(GROUP_FIELDS, line.split(':'))) for line in raw.strip().split('\n')]

            for user in users:
                user[LIST_FIELD] = user[LIST_FIELD].split(',')

            return {user[ID_FIELD]: user for user in users}


def make_shadow_compatible(orig_word):
    """Ensure that orig_word is a valid user/group name for standard shadow utils.

    While this could in theory be achived by simply substituting all non-allowed chars with a valid
    one, we try to translitare sensibly, so that usernames look nicer and to avoid collisions. See
    inline comments for further details.

    Any change made to the word is logged with level WARNING.

    """
    if orig_word is None:
        return None
    # Encode German Umlauts
    word = orig_word.translate(str.maketrans({
        'ä': 'ae', 'ö': 'oe', 'ü': 'ue',
        'Ä': 'Ae', 'Ö': 'Oe', 'Ü': 'Ue',
        'ß': 'ss',
        '!': 'i', '$': 's',
        '*': 'x', '@': '_at_',
    }))

    # Downcase
    word = word.lower()

    # Unicode -> Ascii
    word = unidecode(word)

    # Shadow will das Namen mit Kleinbuchstaben oder Underscore anfangen
    if regex.match(r'^[a-z_]', word):
        word = word
    else:
        if len(word) >= 32:
            word = '_'+word[1:]
        else:
            word = '_' + word

    # Das ist der doofe part. Für die ganzen Sonderzeichen gibt es nicht wirklich
    # eine transliterierung in [-0-9_a-z], daher nehme ich einfach underscore,
    # was ggf. zu Kollisionen führen kann. Witzig: Shadow erlaubt '$' im namen,
    # aber nur *ganz* am Ende ...
    word = regex.sub(r'[^-0-9_a-z]', '_', word[:-1]) + regex.sub(r'[^-0-9_a-z$]', '_', word[-1])


    # usernames and group names can only be 32 characters long.
    # My fix is to remove characters a) after the first '_' if there is one.
    #                                b) from the beginning if there is none
    # Also adds two dots as an indicator for where the shortening took place

    excess_chars = len(word) - 32
    if excess_chars > 0:

        if len(word.split('_')) == 1: # no '_' found:
            word = '__' + word[excess_chars+2:]
            # logger.warning(F"shortened {orig_word} to {word}")

        elif len(word.split('_')) > 1: # at least one '_' found:
            fragments = word.split('_')
            if len(fragments[1]) > excess_chars: # we're fine, we can cut excess chars from fragments alone
                fragments[1]=".."+fragments[1][excess_chars+2:]
                word = '_'.join(fragments)
            else:
                logger.error(F"User or group name is too long: {word} ({len(word)})")
                raise(ValueError)
            # logger.warning(F"shortened {orig_word} to {word}")

    if word != orig_word:
        logger.warning("Name '{}' changed to '{}' for shadow compatibilty".format(orig_word, word))

    return word



"""Simple example backend, for documentation purposes only.
It's in the proof-of-concept state.
"""


import logging

# See https://ldap3.readthedocs.io/en/latest/operations.html
from ldap3 import Server, Connection, ALL, BASE, MODIFY_REPLACE, MODIFY_DELETE

logger = logging.getLogger(__name__)

class User:
    """Manages the user object on the service."""
    def __init__(self, userinfo):
        """
        Arguments:
        userinfo -- (type: UserInfo)
        """
        self.primary_group = Group(userinfo.primary_group)
        self.userinfo = userinfo

        s = Server('localhost', get_info=ALL)  # define an unsecure LDAP server
        self.connection = Connection(s, user='cn=admin,dc=foo,dc=ka,dc=bw-cloud-instance,dc=org', password='foobar123')
        self.connection.bind()

    def exists(self):
        logger.info(F"Uros: exist: {self.userinfo.userinfo['eduperson_targeted_id']}")
        """Return whether the user exists on the service.

        If this returns True,  calling `create` should have no effect or raise an error.
        """
        users_name = self.userinfo.userinfo['eduperson_targeted_id'][0]
        return self.connection.search(
            #f"cn={self.userinfo.username},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
            f"cn={users_name},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
            '(objectClass=*)',
            search_scope=BASE
        )

    def name_taken(self):
        """Return wheter the username is already taken by another user on the service.

        Should return True if the name is available for this user (even if it is unavailable
        for other users for some reason)
        """
        return bool()
    def create(self):
        """Create the user on the service.

        If the user already exists, do nothing or raise an error
        """
        users_name = self.userinfo.userinfo['eduperson_targeted_id'][0]
        self.connection.add(
            #f"cn={self.userinfo.username},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
            f"cn={users_name},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
            attributes={
                #'objectClass':  ['inetOrgPerson'],
                #'objectClass':  ['inetOrgPerson', 'ldapPublicKey'],
                'objectClass':  ['inetOrgPerson', 'ldapPublicKey','eduPerson'],
                'sn': self.userinfo.family_name,
                'givenName': self.userinfo.given_name,
                'mail': self.userinfo.email,
                'eduPersonPrincipalName': self.userinfo.userinfo['eduperson_principal_name'],
                'eduPersonEntitlement': self.userinfo.userinfo['eduperson_entitlement']
                #'sshPublicKey' = self.userinfo.ssh_keys[0]['value'] 
            })

    def update(self):
        """Update all relevant information about the user on the service.

        If the user doesn't exists, behavour is undefined.
        """
        #users_name = self.userinfo.userinfo['eduperson_targeted_id'][0]
        #self.connection.modify(
        #    #f"cn={self.userinfo.username},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
        #    f"cn={users_name},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
        #    {
        #        #'sn': self.userinfo.family_name,
        #        'givenName': [(MODIFY_REPLACE, [self.userinfo.given_name])],
        #        #'mail': self.userinfo.email,
        #        #'sshPublicKey': [(MODIFY_REPLACE, [SSHKey])],
        #        #'eduPersonPrincipalName': [(MODIFY_REPLACE, [self.userinfo.userinfo['eduperson_principal_name']])],
        #        #'eduPersonEntitlement': [(MODIFY_REPLACE, [self.userinfo.userinfo['eduperson_entitlement']])],
        #        }
        #    )
        pass

    def delete(self):
        """Delete the user on the service.

        If the user doesn't exists, do nothing or raise an error.
        """
        users_name = self.userinfo.userinfo['eduperson_targeted_id'][0]
        #self.connection.delete(f"cn={self.userinfo.username},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org")
        self.connection.delete(f"cn={users_name},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org")

    def mod(self, supplementary_groups=None):
        """Modify the user on the service.

        The state of the user with respect to the provided Arguments after calling this function
        should not depend on the state the user had previously.

        If the user doesn't exists, behavour is undefined.

        Arguments:
        supplemantary_groups -- A list of groups to add the user to (type: list(Group))
        """
        #users_name = self.userinfo.userinfo['eduperson_targeted_id'][0]
        #self.connection.modify(
        #    #f"cn={self.userinfo.username},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
        #    f"cn={users_name},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
        #    {
        #        #'sn': self.userinfo.family_name,
        #        'givenName': [(MODIFY_REPLACE, [self.userinfo.given_name])],
        #        #'mail': self.userinfo.email,
        #        #'sshPublicKey': [(MODIFY_REPLACE, [SSHKey])],
        #        #'eduPersonPrincipalName': [(MODIFY_REPLACE, [self.userinfo.userinfo['eduperson_principal_name']])],
        #        #'eduPersonEntitlement': [(MODIFY_REPLACE, [self.userinfo.userinfo['eduperson_entitlement']])],
        #        }
        #    )
        pass
    def install_ssh_keys(self):
        logger.info(F"Uros: key: {self.userinfo.ssh_keys[0]['value']}")
        SSHKey = self.userinfo.ssh_keys[0]['value'] 
        users_name = self.userinfo.userinfo['eduperson_targeted_id'][0]
        self.connection.modify(
            #f"cn={self.userinfo.username},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
            f"cn={users_name},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
            {
                #'sn': self.userinfo.family_name,
                #'givenName': self.userinfo.given_name,
                #'mail': self.userinfo.email,
                'sshPublicKey': [(MODIFY_REPLACE, [SSHKey])],
                }
            )
        """Install users SSH keys on the service.

        No other SSH keys should be active after calling this function.

        If the user doesn't exists, behaviour is undefined.
        """
        pass
    def uninstall_ssh_keys(self):
        logger.info(F"Uros delete: key: {self.userinfo.ssh_keys[0]['value']}")
        SSHKey = self.userinfo.ssh_keys[0]['value'] 
        users_name = self.userinfo.userinfo['eduperson_targeted_id'][0]
        self.connection.modify(
            #f"cn={self.userinfo.username},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
            f"cn={users_name},dc=foo,dc=ka,dc=bw-cloud-instance,dc=org",
            {
                'sshPublicKey': [(MODIFY_DELETE, [SSHKey])],
                }
            )
        """Uninstall the users SSH keys on the service.

        This must uninstall all SSH keys installed with `install_ssh_keys`. It may uninstall SSH
        keys installed by other means.

        If the user doesn't exists, behaviour is undefined.
        """
        pass
    @property
    def credentials(self):
        """Return any additional login information required to access the service after deployment.

        This should not include the installed SSH keys.

        If the user is not fully deployed, behaviour is undefined.
        """
        return {}

class Group:
    """Manages the group object on the service."""
    def __init__(self, name):
        """
        Arguments:
        name -- The name of the group
        """
        self.name = name
    def exists(self):
        """Return whether the group already exists."""
        pass
    def create(self):
        """Create the group on the service.

        If the group already exists, behaviour is undefined.
        """
        pass



"""
Generate useful user or group names
"""

# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
# pylint: disable=raise-missing-from, missing-docstring, too-few-public-methods

import logging
from .config import CONFIG

logger = logging.getLogger(__name__)


class FriendlyNameGenerator():
    """
    Create Names from UserInfo. 
    Don't return the same name twice per run
    """
    dont_use_these_names  = []
    dont_use_these_names.append('[]')
    dont_use_these_names.append('none')
    strategies = [
            "{self.userinfo.preferred_username}",
            "{self.userinfo.given_name}",
            "{self.userinfo.given_name:.3}{self.userinfo.family_name:.3}",
            "{self.userinfo.family_name}",
            "{self.userinfo.given_name:.4}{self.userinfo.family_name:.3}",
            "{self.userinfo.given_name:.5}{self.userinfo.family_name:.3}",
            "{self.userinfo.given_name:.2}{self.userinfo.family_name:.3}",
            "{self.userinfo.given_name:.4}{self.userinfo.family_name:.4}",
            "{self.userinfo.given_name:.5}{self.userinfo.family_name:.4}",
            "{self.userinfo.given_name:.2}{self.userinfo.family_name:.4}",
            "{self.userinfo.given_name:.4}{self.userinfo.family_name:.5}",
            "{self.userinfo.given_name:.5}{self.userinfo.family_name:.5}",
            "{self.userinfo.given_name:.2}{self.userinfo.family_name:.5}",
            "{self.userinfo.given_name:.4}{self.userinfo.family_name:.2}",
            "{self.userinfo.given_name:.5}{self.userinfo.family_name:.2}",
            "{self.userinfo.given_name:.2}{self.userinfo.family_name:.2}",
            "{self.userinfo.email}",
            ]
    next_strategy_idx = -1

    def __init__(self, userinfo):
        """Generate a useful name"""
        self.userinfo = userinfo
    def suggest_name(self, suggestion=None, forbidden_names = None):
        # Copy forbidden names:
        for name in forbidden_names or []:
            if name.lower() not in self.dont_use_these_names:
                self.dont_use_these_names.append(name.lower())

        while True:

            self.next_strategy_idx += 1
            try:
                candidate_name = self.strategies[self.next_strategy_idx].format(**locals()).lower().replace('@','-')
            except KeyError as e:
                pass
                continue
            except AttributeError as e:
                pass
                continue
            except IndexError:
                NL="\n    "
                logger.error(F"Ran out of strategies for generating a friendly username")
                logger.error(F"The list of tried usernames is: \n {NL.join(self.dont_use_these_names)}")
                raise
                return None

            if candidate_name.lower() not in self.dont_use_these_names:
                self.dont_use_these_names.append(candidate_name)
                if CONFIG.getboolean('messages', 'log_username_creation', fallback=False):
                    logger.info(F"Potential username: '{candidate_name}'")
                return candidate_name.lower()
            else:
                self.dont_use_these_names.append(candidate_name.lower())

        return None
    def tried_names(self):
        return self.dont_use_these_names


class PooledNameGenerator():
    """Name Generator for Pooled Accounts"""
    index  = 0
    digits = CONFIG.getint('username_generator','pool_digits', fallback=3)
    username_prefix = ""
    def __init__(self, pool_prefix = 'pool'):
        self.username_prefix = CONFIG.get('username_generator','pool_prefix',
                fallback=pool_prefix) 
        if self.username_prefix is None:
            self.username_prefix = 'pool'

    def suggest_name(self, **kwargs):
        self.index += 1
        candidate_name = F"{self.username_prefix}%0{self.digits}d" % self.index
        return candidate_name



# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=mixed-indentation, wrong-import-order
# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods
import os
import sys

from feudal_globalconfig import globalconfig
from configparser import ConfigParser
from pathlib import Path
import logging

from .results import Failure

PARSE_CMDLINE_PARAMETERS = True
if 'pytest' in sys.modules:
    PARSE_CMDLINE_PARAMETERS = False
else:
    try:
        PARSE_CMDLINE_PARAMETERS = globalconfig.config['parse_commandline_args']
    except KeyError as e:
        pass

if PARSE_CMDLINE_PARAMETERS:
    from ldf_adapter.cmdline_params import args

CONFIG = ConfigParser()


def reload():
    """Reload configuration from disk.

    Config locations, by priority:
    --config option (defaults to /etc/feudal/feudal_adapter.conf)
    $FEUDAL_ADAPTER_CONFIG
    ./feudal_adapter.conf
    ~/.config/feudal/feudal_adapter.conf
    /etc/feudal/feudal_adapter.conf

    processing is stopped, once a give file is found
    """

    logger = logging.getLogger(__name__)
    files = []

    # If the program has arguments with a config_file: prefer it:
    if PARSE_CMDLINE_PARAMETERS:
        files.insert(0, Path(args.config_file))

    # If the caller of the library has provided a configfile: prefer it:
    logger.debug(F"Files: {files}")
    try:
        globalconf_conf_file = Path(globalconfig.config['CONFIGFILE'])
        logger.debug(F"Trying config of globalconfig: {globalconfig.config['CONFIGFILE']}")
        if globalconf_conf_file.exists():
            files.insert(0, globalconf_conf_file)
    except KeyError:
        pass
    
    # Finally, check the environment (last means highes priority)
    filename = os.environ.get("FEUDAL_ADAPTER_CONFIG")
    if filename is None:
        filename = os.environ.get("LDF_ADAPTER_CONFIG")
    if filename:
        files.append(Path(filename))

    # default files
    files += [
        Path('feudal_adapter.conf'),
        Path.home()/'.config'/'feudal_adapter.conf',
        Path.home()/'.config'/'feudal'/'feudal_adapter.conf',
        Path('/etc/feudal/feudal_adapter.conf'),
        Path('ldf_adapter.conf'),
        Path.home()/'.config'/'ldf_adapter.conf',
        Path.home()/'.config'/'feudal'/'ldf_adapter.conf',
        Path('/etc/feudal/ldf_adapter.conf')
    ]

    config_loaded = False
    for f in files:
        if f.exists():
            files_read = CONFIG.read(f)
            logger.debug(F"Using this config file: {files_read}")
            globalconfig.info = {}
            globalconfig.info['config_files_read'] = files_read
            config_loaded = True
            break
    if not config_loaded:
        logger.warning("Could not find any config file")
        logger.debug("Trying to copy config from globalconfig")
        logger.debug(F"type of CONFIG: {type(CONFIG)}")
        # try:
        #     logger.debug(F"type of globalconfig.config['CONFIG']: {type(globalconfig.config['CONFIG'])}")
        # except KeyError:
        # raise Failure(message="Could not find any config (neither file not in globalconfig)")
        logger.error("Could not find any config (neither file not in globalconfig")
        exit(4)

# Load config on import
reload()
